"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020  holukas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import datetime as dt
import fnmatch
import logging
import os
import shutil
import sys
import time
from pathlib import Path

import pandas as pd


def set_dirs(indir: False or Path,
             outdir: False or Path):
    """
    In case no input or output directory is given, use folders 'input'
    and 'output' in current working directory.

    Returns
    -------
    Input and output directories as Path

    """
    cwd = Path(os.path.dirname(os.path.abspath(__file__)))  # Current working directory
    if not indir:
        indir = cwd / 'input'
    if not outdir:
        outdir = cwd / 'output'
    return indir, outdir


def create_logger(name: str, logfile_path: Path = None):
    """
    Create name logger and log outputs to file

    A new logger is only created if name does not exist.

    Parameters
    ----------
    logfile_path: Path
        Path to the log file to which the log output is saved.
    name:
        Corresponds to the __name__ of the calling file.

    Returns
    -------
    logger class

    References
    ----------
    https://www.youtube.com/watch?v=jxmzY9soFXg
    https://stackoverflow.com/questions/53129716/how-to-check-if-a-logger-exists
    """

    logger = logging.getLogger(name)

    # Only create new logger if it does not exist already for the respective module,
    # otherwise the log output would be printed x times because a new logger is
    # created everytime the respective module is called.
    if not logger.hasHandlers():
        logger.setLevel(logging.DEBUG)

        formatter = logging.Formatter('%(asctime)s:%(name)s:  %(message)s')

        file_handler = logging.FileHandler(logfile_path)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)

        stream_handler = logging.StreamHandler(stream=sys.stdout)
        stream_handler.setFormatter(formatter)

        logger.addHandler(file_handler)
        logger.addHandler(stream_handler)

    return logger


class CreateOutputDirs:
    def __init__(self, dyco_instance):
        self.root_dir = dyco_instance.outdir
        self.del_previous_results = dyco_instance.del_previous_results
        self.phase = dyco_instance.phase
        self.phase_files = dyco_instance.phase_files

    def required_dirs(self):
        outdirs = [
            '0-0_log',
            f'{self.phase}-0_{self.phase_files}_overview',
            f'{self.phase}-1_{self.phase_files}_covariances',
            f'{self.phase}-2_{self.phase_files}_covariances_plots',
            f'{self.phase}-3_{self.phase_files}_time_lags_overview',
            f'{self.phase}-4_{self.phase_files}_time_lags_overview_histograms',
            f'{self.phase}-5_{self.phase_files}_time_lags_overview_timeseries',
            f'{self.phase}-7_{self.phase_files}_normalized',
        ]

        if self.phase != 3:
            outdirs.append(f'{self.phase}-6_{self.phase_files}_normalization_lookup_table')
        else:
            outdirs.append(f'{self.phase}-6_{self.phase_files}_final_time_lags_lookup_table')
        return outdirs

    def setup_output_dirs(self):
        """Make output directories."""
        required_dirs = self.required_dirs()
        outdirs = {}

        # # Delete previous results
        # if self.del_previous_results and os.path.isdir(self.root_dir):
        #     print(f"Deleting folder {self.root_dir} ...")
        #     shutil.rmtree(self.root_dir)

        # Create Path to required directories, store keys and full paths in dict
        for nd in required_dirs:
            outdirs[nd] = self.root_dir / nd

        # Delete dirs if needed
        for key, path in outdirs.items():
            if Path.is_dir(path) and self.del_previous_results:
                if path.stem == '0-0_log':
                    continue
                print(f"Deleting folder {path} ...")
                shutil.rmtree(path)

        # Make dirs
        for key, path in outdirs.items():
            if not Path.is_dir(path):
                print(f"Creating folder {path} ...")
                os.makedirs(path)

        # for filename in os.listdir(path):
        #     filepath = os.path.join(path, filename)
        #     try:
        #         if os.path.isfile(filepath) or os.path.islink(filepath):
        #             print(f"Deleting file {filepath} ...")
        #             os.unlink(filepath)
        #         # elif os.path.isdir(filepath):
        #         #     shutil.rmtree(filepath)
        #     except Exception as e:
        #         print('Failed to delete %s. Reason: %s' % (filepath, e))

        return outdirs


class FilesDetector:
    """
    Create overview dataframe of available and missing (expected) files
    """
    found_files = []
    files_overview_df = pd.DataFrame()

    def __init__(self,
                 dyco_instance,
                 logfile_path: Path = None,
                 outdir: bool or Path = False):
        """
        Parameters
        ----------
        outdir: Path or False
            Export folder to save the files overview to.
        """

        self.indir = dyco_instance.indir
        self.fnm_pattern = dyco_instance.fnm_pattern
        self.fnm_date_format = dyco_instance.fnm_date_format
        self.file_generation_res = dyco_instance.file_generation_res
        self.dat_recs_nominal_timeres = dyco_instance.dat_recs_nominal_timeres
        self.files_how_many = dyco_instance.files_how_many
        self.outdir = outdir

        self.logger = create_logger(logfile_path=logfile_path, name=__name__)

    def run(self):
        """Execute processing stack"""
        self.logger.info("Start file search")
        self.found_files = self.search_files(indir=self.indir,
                                             fnm_pattern=self.fnm_pattern)
        if not self.found_files:
            self.logger.error(f"\n(!)ERROR No files found with pattern {self.fnm_pattern}. Stopping script.")
            sys.exit()

        self.files_overview_df = self.add_expected()
        self.files_overview_df = self.add_unexpected()
        self.files_overview_df = self.calc_expected_values()
        self.files_overview_df.loc[:, 'file_available'].fillna(0, inplace=True)
        self.files_overview_df = self.limit_num_files()
        self.logger.info(f"Found {int(self.files_overview_df['file_available'].sum())} available files")

        if self.outdir:
            self.export()

    def get(self):
        return self.files_overview_df

    def export(self):
        """Export dataframe to csv."""
        outfile = '0_files_overview.csv'
        outpath = self.outdir / outfile
        self.files_overview_df.to_csv(outpath)
        self.logger.info(f"Exported file {outfile}")

    @staticmethod
    def search_files(indir, fnm_pattern):
        """
        Search files in indir

        Parameters
        ----------
        indir: Path
            Directory that will be searched.
        fnm_pattern: str
            Filename search pattern, accepts regex.

        Returns
        -------
        List of found files
        """

        found_files = []
        for root, dirs, files in os.walk(str(indir)):
            root = Path(root)
            for idx, filename in enumerate(files):
                if fnmatch.fnmatch(filename, fnm_pattern):
                    filepath = Path(root) / Path(filename)
                    found_files.append(filepath)
        found_files.sort()  # Sorts inplace
        return found_files

    def add_expected(self):
        """
        Create index of expected files (regular start time) and check
        which of these regular files are available

        Returns
        -------
        pandas Dataframe
        """

        # len_before = len(self.found_files)
        first_file_dt = dt.datetime.strptime(self.found_files[0].stem, self.fnm_date_format)
        last_file_dt = dt.datetime.strptime(self.found_files[-1].stem, self.fnm_date_format)
        expected_end_dt = last_file_dt + pd.Timedelta(self.file_generation_res)
        expected_index_dt = pd.date_range(first_file_dt, expected_end_dt, freq=self.file_generation_res)
        files_df = pd.DataFrame(index=expected_index_dt)

        for file_idx, filepath in enumerate(self.found_files):
            filename = filepath.stem
            file_start_dt = dt.datetime.strptime(filename, self.fnm_date_format)

            if file_start_dt in files_df.index:
                files_df.loc[file_start_dt, 'file_available'] = 1
                files_df.loc[file_start_dt, 'filename'] = filename
                files_df.loc[file_start_dt, 'start'] = file_start_dt
                files_df.loc[file_start_dt, 'filepath'] = filepath
                files_df.loc[file_start_dt, 'filesize'] = Path(filepath).stat().st_size
                # files_df.loc[file_start_dt, 'expected_file'] = file_start_dt

        files_df.insert(0, 'expected_file', files_df.index)  # inplace
        return files_df

    def add_unexpected(self):
        """
        Add info about unexpected files (irregular start time)

        Returns
        -------
        pandas DataFrame with added info about irregular files
        """
        files_df = self.files_overview_df.copy()
        for file_idx, filepath in enumerate(self.found_files):
            filename = filepath.stem
            file_start_dt = dt.datetime.strptime(filename, self.fnm_date_format)

            if file_start_dt not in files_df.index:
                files_df.loc[file_start_dt, 'file_available'] = 1
                files_df.loc[file_start_dt, 'filename'] = filename
                files_df.loc[file_start_dt, 'start'] = file_start_dt
                files_df.loc[file_start_dt, 'filepath'] = filepath
                files_df.loc[file_start_dt, 'filesize'] = Path(filepath).stat().st_size

        files_df.sort_index(inplace=True)
        return files_df

    def calc_expected_values(self):
        """
        Calculate expected end time, duration and number of records for each file

        Returns
        -------
        pandas DataFrame with added info about expected values
        """
        files_df = self.files_overview_df.copy()
        files_df['expected_end'] = files_df.index
        files_df['expected_end'] = files_df['expected_end'].shift(-1)
        files_df['expected_duration'] = (files_df['expected_end'] - files_df['start']).dt.total_seconds()
        files_df['expected_records'] = files_df['expected_duration'] / self.dat_recs_nominal_timeres
        # files_df['expected_end'] = files_df['start'] + pd.Timedelta(file_generation_res)
        # files_df.loc[files_df['file_available'] == 1, 'next_file'] = files_df['expected_file']
        # files_df['next_file'] = files_df['next_file'].shift(-1)
        return files_df

    def limit_num_files(self):
        """
        Limit the number of files used

        Returns
        -------
        pandas DataFrame
        """
        if self.files_how_many:
            self.logger.info(f"Limit number of files to {int(self.files_how_many)}")
            for idx, file in self.files_overview_df.iterrows():
                _df = self.files_overview_df.loc[self.files_overview_df.index[0]:idx]
                num_available_files = _df['file_available'].sum()
                if num_available_files >= self.files_how_many:
                    files_overview_df = _df.copy()
                    break

        files_available = self.files_overview_df['file_available'].sum()
        if files_available == 0:
            msg = f"No files available, stopping script (files_available = {files_available})"
            self.logger.critical(msg)
            sys.exit()

        return self.files_overview_df


def generate_run_id():
    """Generate unique id for this run"""
    script_start_time = time.strftime("%Y-%m-%d %H:%M:%S")
    run_id = time.strftime("%Y%m%d-%H%M%S")
    run_id = f"DYCO-{run_id}"
    return run_id, script_start_time


def set_logfile_path(run_id: str, outdir: Path, phase: int):
    """Set full path to log file"""
    name = f'{run_id}.log'
    # name = f'{run_id}_phase-{phase}.log'
    path = outdir / name
    return path



"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020  holukas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import os
from pathlib import Path

import numpy as np
import pandas as pd

from . import setup_dyco, files, loop, plot
from .analyze import AnalyzeLags
from .correction import RemoveLags

pd.set_option('display.max_columns', 15)
pd.set_option('display.width', 1000)


def dyco(cls):
    """
    Wrapper function for DYCO processing chain

    Parameters
    ----------
    cls: class
        The class that is wrapped.

    Returns
    -------
    Wrapper

    """

    class ProcessingChain:

        def __init__(self, **args):
            # PHASE 1 - First normalization to default lag
            # ============================================
            args['phase'] = 1
            self.run_phase_1_input_files = cls(**args)

            # PHASE 2 - Second normalization to default lag
            # =============================================
            args['phase'] = 2
            args = self._update_args(args=args,
                                     prev_phase=self.run_phase_1_input_files.phase,
                                     prev_phase_files=self.run_phase_1_input_files.phase_files,
                                     prev_outdir_files=self.run_phase_1_input_files.outdir,
                                     prev_last_iteration=self.run_phase_1_input_files.lgs_num_iter,
                                     prev_outdirs=self.run_phase_1_input_files.outdirs)
            self.run_phase_2_normalized_files = cls(**args)

            # PHASE 3 - Correction for instantaneous lag
            # ==========================================
            args['phase'] = 3
            args = self._update_args(args=args,
                                     prev_phase=self.run_phase_2_normalized_files.phase,
                                     prev_phase_files=self.run_phase_2_normalized_files.phase_files,
                                     prev_outdir_files=self.run_phase_2_normalized_files.outdir,
                                     prev_last_iteration=self.run_phase_2_normalized_files.lgs_num_iter,
                                     prev_outdirs=self.run_phase_2_normalized_files.outdirs)
            self.run_phase_3_finalize = cls(**args)

            # FINALIZE - Make some more plots summarizing Phases 1-3
            # ======================================================
            plot.SummaryPlots(instance_phase_1=self.run_phase_1_input_files,
                              instance_phase_2=self.run_phase_2_normalized_files,
                              instance_phase_3=self.run_phase_3_finalize)

        def _update_args(self, args, prev_phase, prev_phase_files, prev_outdir_files, prev_last_iteration,
                         prev_outdirs):
            """Update args for running Phases 2 and 3: use results from Phase 1 and 2, respectively"""
            if args['phase'] == 2:
                args['lgs_winsize'] = self._update_winsize(prev_phase=prev_phase,
                                                           prev_phase_files=prev_phase_files,
                                                           prev_last_iteration=prev_last_iteration,
                                                           prev_outdirs=prev_outdirs)
            else:
                args['lgs_winsize'] = 100  # Small window for instantaneous search in Phase 3
                args['lgs_num_iter'] = 1

            args['indir'] = prev_outdir_files / f"{prev_phase}-7_{prev_phase_files}_normalized"
            args['var_lagged'] = f"{args['var_lagged']}_DYCO"  # Use normalized signal
            filename, file_extension = os.path.splitext(args['fnm_pattern'])
            args['fnm_pattern'] = f"{filename}_DYCO{file_extension}"  # Search normalized files
            args['fnm_date_format'] = f"{args['fnm_date_format']}_DYCO"  # Parse file names of normalized files
            var_target = [var_target + '_DYCO' for var_target in args['var_target']]  # Use normalized target cols
            args['var_target'] = var_target
            return args

        def _update_winsize(self, prev_phase, prev_phase_files, prev_last_iteration, prev_outdirs):
            """
            Calculate the range of the lag search window from the last iteration and use
            it for lag search in normalized files

            During the last iteration, this window was detected as the *next* window for
            the *next* iteration, i.e. it was not yet used for lag detection.

            Returns
            -------
            Time window for lag search in Phase 2 iteration 1 and in Phase 3
            """
            filepath_last_iteration = \
                prev_outdirs[
                    f"{prev_phase}-3_{prev_phase_files}_time_lags_overview"] \
                / f'{prev_last_iteration}_segments_found_lag_times_after_iteration-{prev_last_iteration}.csv'

            segment_lagtimes_last_iteration_df = \
                files.read_segment_lagtimes_file(filepath=filepath_last_iteration)
            lgs_winsize = \
                [segment_lagtimes_last_iteration_df['lagsearch_next_start'].unique()[0],
                 segment_lagtimes_last_iteration_df['lagsearch_next_end'].unique()[0]]

            lgs_winsize_normalized = np.abs(lgs_winsize[0] - lgs_winsize[1])  # Range
            lgs_winsize_normalized = lgs_winsize_normalized / 2  # Normalized search window +/- around zero
            return lgs_winsize_normalized

    return ProcessingChain


@dyco
class DynamicLagCompensation:
    """
    DYCO - Dynamic lag compensation
    """

    files_overview_df = pd.DataFrame()

    def __init__(self,

                 var_reference: str,
                 var_lagged: str,
                 phase: int = 1,
                 phase_files: str = 'input_files',
                 fnm_date_format: str = '%Y%m%d%H%M%S',
                 del_previous_results: bool = False,
                 fnm_pattern: str = '*.csv',
                 dat_recs_timestamp_format: False or str = False,
                 files_how_many: False or int = False,
                 file_generation_res: str = '30T',
                 file_duration: str = '30T',
                 lgs_segment_dur: False or pd.DateOffset = False,
                 lgs_hist_perc_thres: float = 0.9,
                 lgs_hist_remove_fringe_bins: bool = True,
                 dat_recs_nominal_timeres: float = 0.05,
                 lgs_winsize: int = 1000,
                 lgs_num_iter: int = 3,
                 indir: Path = False,
                 outdir: Path = False,
                 target_lag: int = 0,
                 var_target: list = None):
        """

        Parameters
        ----------
        phase: int
            Phase in the processing chain, automatically filled attributed during processing.
            * Phase 1 works on input files and applies the first normalization.
            * Phase 2 works on normalized files from Phase 1 and refines normalization.

        var_reference: str
            Column name of the reference signal in the data. Lags are
            determined in relation to this signal.

        var_lagged: str
            Column name of the lagged signal  for which the lag time in
            relation to the reference signal is determined.

        fnm_pattern: str, accepts regex
            Filename pattern for data file search.
            Example:
                - With data files following the naming structure '20161015123000.csv'
                the corresponding setting is: fnm_pattern='2016*.csv'

        fnm_date_format: str
            Date format in data filenames. Is used to parse the date and
            time info from the filename of found files. Only files found
            with *fnm_pattern* will be parsed.
            Example:
                - With a data file named '20161015123000.csv' the
                 corresponding setting is: fnm_date_format='%Y%m%d%H%M%S'


        del_previous_results: bool
            If *True*, delete all previous results in *indir*. If *False*,
            search for previously calculated results and continue.

        dat_recs_timestamp_format: str
            Timestamp format for each row record.

        files_how_many: int
            Limits number of found files that are used.

        file_generation_res: str (pandas DateOffset)
            Frequency at which new files were generated. This does not
            relate to the data records but to the file creation time.
            Examples:
                * '30T' means a new file was generated every 30 minutes.
                * '1H' means a new file was generated every hour.
                * '6H' means a new file was generated every six hours.
            For pandas DateOffset options see:
                https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases

        file_duration: str (pandas DateOffset)
            Duration of one data file.
            Example:
                * '30T': data file contains data from 30 minutes.
            For pandas DateOffset options see:
                https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases

        lgs_segment_dur: str (pandas DateOffset)
            Segment duration for lag determination. If it is the same
            as *file_duration*, the lag time for the complete file is
            calculated from all file data. If it is shorter than
            *file_duration*, then the file data is split into segments
            and the lag time is calculated for each segment separately.
            Examples:
                * '10T': calculates lag times for 10-minute segments.
                * With the settings
                    file_duration = '30T' and
                    lgs_segments_dur = '10T'
                    the 30-minute data file is split into three 10-minute
                    segments and the lag time is determined in each of the
                    segments, yielding three lag times.
            For pandas DateOffset options see:
                https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases

        lgs_hist_perc_thres: float between 0.1 and 1 (percentage)
            Cumulative percentage threshold in histogram of found lag times.
            The time window for lag search during each iteration (i) is
            narrowed down based on the histogram of all found lag times
            from the previous iteration (i-1). Is set to 1 if > 1 or set
            to 0.1 if < 0.1.

            During each iteration and after lag times were determined for
            all files and segments, a histogram of found lag times is created
            to identify the histogram bin in which most lag times (most counts)
            were found (peak bin). To narrow down the time window for lag search
            during the next iteration (i+1), the bins around the peak bin are
            included until a certain percentage of the total values is reached
            over the expanded bin range, centered on the peak bin.

            Example:
                * 0.9: include all bins to each site of the peak bin until 90%
                    of the total found lag times (counts) are included. The time
                    window for the lag search during the next iteration (i+1) is
                    determined by checking the left side (start) of the first
                    included bin and the right side (end) of the last included
                    bin.

        lgs_hist_remove_fringe_bins: bool
            Remove fringe bins in histogram of found lag times. In case of low
            signal-to-noise ratios the lag search yields less clear results and
            found lag times tend to accumulate in the fringe bins of the histogram,
            i.e. in the very first and last bins, potentially creating non-desirable
            peak bins. In other words, if True the first and last bins of the
            histogram are removed before the time window for lag search is adjusted.

        dat_recs_nominal_timeres: float
            Nominal (expected) time resolution of data records.
            Example:
                * 0.05: one record every 0.05 seconds (20Hz)

        lgs_winsize: int
            Starting time window size for lag search +/-, given as number of records.
            If negative, the absolute value will be used.
            Example:
                * 1000: Lag search during the first iteration is done in a time window
                    from -1000 records to +1000 records.

        lgs_num_iter: int
            Number of lag search interations. Before each iteration, the time window
            for the lag search is narrowed down, taking into account results from the
            previous iteration. Exception is the first iteration for which the time
            window as given in *lgs_winsize* is used.
            Example:
                * *lgs_num_iter* = 3: lag search in iteration 1 (i1) uses *lgs_winsize*
                    to search for lag times, then the lag window is narrowed down using
                    results from i1. The adjusted search window is the used in i2 to
                    again search lag times for the same data. Likewise, i3 uses the
                    adjusted search window based on results from i2.

        indir: Path or False
            Source folder that contains the data files. If *False*, a folder named 'input'
            is searched in the current working directory.

        outdir: Path or False
            Output folder for results. If *False*, a folder named 'output'
            is created in the current working directory.

        target_lag: int
            The target lag given in records to which lag times of all files are
            normalized. A negative number means that *var_lagged* lags x records
            behind *var_reference*.
            Example:
                * 0: The default lag time for all files is set to 0 records.
                    This means that if a lag search is performed on these date, the
                    lag time should consistently be found around 0 records.

        var_target: list of strings
            Column names of the time series the normalized lag should be applied to.


        Links
        -----
        * Overview of pandas DateOffsets:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        """

        self.run_id, self.script_start_time = setup_dyco.generate_run_id()

        # Setup for Phases 1-3
        self.phase = phase
        if self.phase == 1:
            self.phase_files = '_input_files_'
        elif self.phase == 2:
            self.phase_files = '_normalized_files_'
        else:
            self.phase_files = '_refined_normalized_files_'

        # Input and output directories
        self.indir, self.outdir = setup_dyco.set_dirs(indir=indir, outdir=outdir)

        # File settings
        self.fnm_date_format = fnm_date_format
        self.file_generation_res = file_generation_res
        self.fnm_pattern = fnm_pattern
        self.files_how_many = files_how_many
        self.file_duration = file_duration
        self.del_previous_results = del_previous_results

        # Data records
        self.dat_recs_nominal_timeres = dat_recs_nominal_timeres
        self.dat_recs_timestamp_format = dat_recs_timestamp_format
        # self.data_segment_overhang = data_segment_overhang

        # Lag search and normalization
        self.var_reference = var_reference
        self.var_lagged = var_lagged
        self.lgs_segment_dur = lgs_segment_dur
        self.lgs_winsize = [abs(lgs_winsize) * -1, abs(lgs_winsize)]
        self.lgs_winsize_initial = self.lgs_winsize
        self.lgs_num_iter = lgs_num_iter
        self.lgs_hist_remove_fringe_bins = lgs_hist_remove_fringe_bins
        if lgs_hist_perc_thres > 1:
            self.lgs_hist_perc_thres = 1
        elif lgs_hist_perc_thres < 0.1:
            self.lgs_hist_perc_thres = 0.1  # Minimum 10% since less would not be useful
        else:
            self.lgs_hist_perc_thres = lgs_hist_perc_thres
        self.target_lag = target_lag
        self.var_target = var_target

        # Indicate if any new iteration data was created, by default *False*.
        # Will be set to *True* automatically when a new iteration is run.
        self.new_iteration_data = False

        # Start scripts
        self.run()

    def run(self):
        """
        Run setup, calculations, analyses and correction of files

        Processing consists of 4 steps:
            * Step 1: Setup
            * Step 2: Calculate time lags: lag times for each file segment
            * Step 3: Analyze time lags: analyze results and create default-lag lookup-table (LUT)
            * Step 4: Remove time lags: use look-up table to normalize time lags across files

        Each step uses results from the previous step.

        """
        # Step 1: Setup
        self.logfile_path, self.files_overview_df = self.setup()

        # Step 2: Calculation of lag times for each file segment in input files
        self.calculate_lags()

        # Step 3: Analyses of results, create LUT
        lut_success = self.analyze_lags()

        # Step 4: Lag-time normalization for each file
        self.remove_lags(lut_success=lut_success)

    def setup(self):
        """Create output folders, start logger and search for files"""
        # Create folders
        self.outdirs = setup_dyco.CreateOutputDirs(dyco_instance=self).setup_output_dirs()

        # Start logging
        logfile_path = setup_dyco.set_logfile_path(run_id=self.run_id,
                                                   outdir=self.outdirs['0-0_log'],
                                                   phase=self.phase)
        logger = setup_dyco.create_logger(logfile_path=logfile_path, name=__name__)
        logger.info(f"Run ID: {self.run_id}")

        # Search files
        fd = setup_dyco.FilesDetector(dyco_instance=self,
                                      outdir=self.outdirs[f'{self.phase}-0_{self.phase_files}_overview'],
                                      logfile_path=logfile_path)
        fd.run()
        files_overview_df = fd.get()
        return logfile_path, files_overview_df

    def calculate_lags(self):
        """
        Calculate covariances and detect covariance peaks to determine lags
        for each file segment
        """
        for iteration in range(1, 1 + self.lgs_num_iter):
            loop_iter = loop.Loop(dyco_instance=self,
                                  iteration=iteration)
            loop_iter.run()
            self.lgs_winsize, self.new_iteration_data = loop_iter.get()  # Update search window for next iteration

        # Plot loop results after all iterations finished TODO ACT
        loop_plots = loop.PlotLoopResults(dyco_instance=self,
                                          plot_cov_collection=True,
                                          plot_hist=True,
                                          plot_timeseries_segment_lagtimes=True)
        loop_plots.run()
        return

    def analyze_lags(self):
        """Analyze lag search results and create look-up table for lag-time normalization"""
        analyze = AnalyzeLags(dyco_instance=self)
        return analyze.lut_available

    def remove_lags(self, lut_success):
        """
        Apply look-up table to normalize lag for each file
        """
        if lut_success:
            RemoveLags(dyco_instance=self)
        return

# if __name__ == "__main__":
#     DynamicLagRemover()



"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020  holukas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import os
from pathlib import Path

import matplotlib.dates as mdates
import matplotlib.gridspec as gridspec
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from . import files, lag, plot, setup_dyco


class Loop:
    """
    Loop through all files and their segments
    """

    def __init__(self,
                 dyco_instance,
                 iteration: int = 1):
        """

        Parameters
        ----------
        dyco_instance:

        iteration: int
            Number of current iteration.
        """
        self.dat_recs_timestamp_format = dyco_instance.dat_recs_timestamp_format
        self.dat_recs_nominal_timeres = dyco_instance.dat_recs_nominal_timeres
        self.lgs_hist_remove_fringe_bins = dyco_instance.lgs_hist_remove_fringe_bins
        self.lgs_hist_perc_thres = dyco_instance.lgs_hist_perc_thres
        self.outdirs = dyco_instance.outdirs
        self.lgs_segment_dur = dyco_instance.lgs_segment_dur
        self.var_reference = dyco_instance.var_reference
        self.var_lagged = dyco_instance.var_lagged
        self.lgs_num_iter = dyco_instance.lgs_num_iter
        self.files_overview_df = dyco_instance.files_overview_df
        self.logfile_path = dyco_instance.logfile_path
        self.lgs_winsize = dyco_instance.lgs_winsize
        self.fnm_date_format = dyco_instance.fnm_date_format
        self.phase = dyco_instance.phase
        self.phase_files = dyco_instance.phase_files
        self.new_iteration_data = dyco_instance.new_iteration_data  # Indicates if new iteration was executed
        # # force_min_stepsize = True if self.iteration == self.lgs_num_iter else False
        # self.shift_stepsize, \
        # self.hist_bin_range = self.lagsearch_settings(win_lagsearch=self.lgs_winsize,
        #                                               force_min_stepsize=False)

        self.iteration = iteration

        self.logger = setup_dyco.create_logger(logfile_path=self.logfile_path, name=__name__)

    def run(self):
        """Loop through all found files"""

        self.logger.info(f"Start FILE LOOP - ITERATION {self.iteration} {'-' * 40}")

        # Use results from previous iteration, if available
        prev_results = self.check_for_prev_results(iteration=self.iteration)
        if not prev_results:
            return None
        else:
            segment_lagtimes_df = prev_results[0]
            self.lgs_winsize = prev_results[1]
            self.new_iteration_data = True  # New iteration output will be generated

        self.shift_stepsize, \
        self.hist_bin_range = self.lagsearch_settings(lgs_winsize=self.lgs_winsize,
                                                      force_min_stepsize=False)

        # data_collection_df = pd.DataFrame() # todo (maybe)

        # Loop files
        for file_idx, file_info_row in self.files_overview_df.iterrows():

            # Check file availability
            if file_info_row['file_available'] == 0:
                continue

            # Read and prepare data file
            data_df = files.read_raw_data(filepath=file_info_row['filepath'],
                                          data_timestamp_format=self.dat_recs_timestamp_format)  # nrows for testing

            # Insert timestamp if missing
            if self.dat_recs_timestamp_format:
                true_resolution = files.calc_true_resolution(num_records=len(data_df),
                                                             data_nominal_res=self.dat_recs_nominal_timeres,
                                                             expected_records=file_info_row['expected_records'],
                                                             expected_duration=file_info_row['expected_duration'])
            else:
                data_df, true_resolution = files.insert_timestamp(df=data_df,
                                                                  file_info_row=file_info_row,
                                                                  num_records=len(data_df),
                                                                  data_nominal_res=self.dat_recs_nominal_timeres,
                                                                  expected_records=file_info_row['expected_records'],
                                                                  expected_duration=file_info_row['expected_duration'])

            # Add raw data info to overview when files are first read (during iteration 1)
            if self.iteration == 1:
                self.files_overview_df = files.add_data_stats(df=data_df,
                                                              true_resolution=true_resolution,
                                                              filename=file_info_row['filename'],
                                                              files_overview_df=self.files_overview_df,
                                                              found_records=len(data_df),
                                                              fnm_date_format=self.fnm_date_format)

            # Loop through data segments in file
            segment_lagtimes_df = self.loop_segments(data_df=data_df,
                                                     file_idx=file_idx,
                                                     filename=file_info_row['filename'],
                                                     segment_lagtimes_df=segment_lagtimes_df)

            # # Collect all file data # todo (maybe)
            # data_collection_df = self.collect_file_data(data_df=data_df,
            #                                             file_idx=file_idx,
            #                                             data_collection_df=data_collection_df)

            # segment_lagtimes_df = lag.calc_quantiles(df=segment_lagtimes_df)

        # Set new search window
        hist_series = self.filter_series(filter_col='iteration', filter_equal_to=self.iteration,
                                         df=segment_lagtimes_df, series_col='PEAK-COVABSMAX_SHIFT')
        lgs_winsize_adj = lag.AdjustLagsearchWindow(series=hist_series,
                                                    outdir=self.outdirs[
                                                        f'{self.phase}-4_{self.phase_files}_time_lags_overview_histograms'],
                                                    iteration=self.iteration,
                                                    plot=True,
                                                    remove_fringe_bins=self.lgs_hist_remove_fringe_bins,
                                                    hist_num_bins=self.hist_bin_range,
                                                    perc_threshold=self.lgs_hist_perc_thres).get()
        self.lgs_winsize = lgs_winsize_adj  # Update lag search window to new range

        # Add time window start and end for next iteration lagsearch
        self.add_lagsearch_adj_info(segment_lagtimes_df=segment_lagtimes_df,
                                    iteration=self.iteration,
                                    next_lgs_winsize=lgs_winsize_adj)

        # plot.lagsearch_collection(lagsearch_df_collection=lagsearch_df_collection,
        #                           outdir=self.outdirs['1-1_PLOTS_segment_covariance'])

        # Plot all found segment lag times
        self.plot_segment_lagtimes_ts(segment_lagtimes_df=segment_lagtimes_df,
                                      outdir=self.outdirs[
                                          f'{self.phase}-5_{self.phase_files}_time_lags_overview_timeseries'],
                                      iteration=self.iteration,
                                      phase=self.phase)

    @staticmethod
    def plot_segment_lagtimes_ts(segment_lagtimes_df: pd.DataFrame,
                                 outdir,
                                 phase: int = 1,
                                 iteration: int = 1,
                                 show_all: bool = False,
                                 overlay_default: bool = False,
                                 overlay_default_df: pd.DataFrame = None,
                                 overlay_target_val: False or int = False):
        """
        Plot time series of found segment lag times

        Parameters
        ----------
        segment_lagtimes_df: pandas DataFrame
            Contains found lag times for all segments.
        outdir
        iteration: int
            Current lag search iteration.
        show_all: bool
            If False, plots only data from current iteration. In this case,
            segment_lagtimes_df is filtered to contain data from the current
            iteration only, other iterations are ignored.
            If True, plots data from all iterations in one plot.
        overlay_default: bool
            Option to plot default lag times for each day.
            If True, plots the 'median' column of overlay_default_df in addition
            to the iteration results.
        overlay_default_df: pandas DataFrame
            Contains the median default lag calculated from the last iteration
            in column 'median'.
        overlay_target_val: False or int
            Draws a horizontal line of the target lag. All files are normalized
            in a way so that the found default lag for each day corresponds to
            the target lag.

        Returns
        -------

        """
        _df = segment_lagtimes_df.copy()
        if not show_all:
            _df = _df.loc[_df['iteration'] == iteration, :]
            txt_info = f"PHASE {phase}, ITERATION {iteration}: found time lags"
            outfile = f"PHASE-{phase}_ITERATION-{iteration}_TIMESERIES-PLOT_segment_lag_times_iteration"
        else:
            txt_info = f"Found lag times from ALL iterations"
            outfile = f"TIMESERIES-PLOT_segment_lag_times_FINAL"

        gs, fig, ax = plot.setup_fig_ax()
        cmap = plt.cm.get_cmap('rainbow', iteration)
        colors = cmap(np.linspace(0, 1, iteration))

        alpha = .5
        iteration_grouped = _df.groupby('iteration')
        for idx, group_df in iteration_grouped:
            ax.plot_date(pd.to_datetime(group_df['start']), group_df['PEAK-COVABSMAX_SHIFT'],
                         alpha=alpha, marker='o', ms=6, color=colors[int(idx - 1)], lw=0, ls='-',
                         label=f'found lag times in iteration {int(idx)}', markeredgecolor='None')

        plot.default_format(ax=ax, label_color='black', fontsize=12,
                            txt_xlabel='segment date', txt_ylabel='lag', txt_ylabel_units='[records]')

        if overlay_default:
            Loop._add_overlay(overlay_default_df=overlay_default_df, ax=ax)

        # if overlay_target_val:
        #     ax.axhline(overlay_target_val, color='black', ls='--', label='target: normalized default lag')

        txt_info = f"PHASE {phase}, ITERATION {iteration}: FOUND TIME LAGS"
        font = {'family': 'sans-serif', 'color': 'black', 'weight': 'bold', 'size': 20, 'alpha': 1, }
        ax.set_title(txt_info, fontdict=font)

        ax.legend(frameon=True, loc='upper right').set_zorder(100)

        # Automatic tick locations and formats
        locator = mdates.AutoDateLocator(minticks=5, maxticks=20)
        formatter = mdates.ConciseDateFormatter(locator)
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter(formatter)

        # Save
        outpath = outdir / outfile
        # print(f"Saving time series of found segment lag times in {outpath} ...")
        fig.savefig(f"{outpath}.png", format='png', bbox_inches='tight', facecolor='w',
                    transparent=True, dpi=150)

        return outfile

    @staticmethod
    def _add_overlay(overlay_default_df, ax):
        if not overlay_default_df.empty:
            # Plot horizontal broken bars
            ax_yrange = np.abs(ax.get_ylim()).sum()
            yrange_offset = (ax_yrange / 50) / 2  # Height of bars
            make_label = True
            for current_date, row_data in overlay_default_df.iterrows():
                current_date = pd.to_datetime(current_date)
                width = pd.Timedelta(days=1)
                xrange = (current_date, width)
                yrange = (row_data['median'] - yrange_offset, yrange_offset * 2)
                yrange = sorted(yrange)
                yrange_target = (row_data['target_lag'] - yrange_offset, yrange_offset * 2)
                if make_label:
                    # Found lag times
                    ax.broken_barh([xrange], yrange, facecolors='#FDD835', alpha=1, edgecolor='None',
                                   label='3-day median lag time (centered)\nfrom high-quality covariance peaks')

                    # Lag times after normalization
                    ax.broken_barh([xrange], yrange_target, facecolors='#8BC34A', alpha=1, edgecolor='None',
                                   label='daily default lag after normalization')

                    # Correction arrow
                    # https://matplotlib.org/api/_as_gen/matplotlib.patches.FancyArrowPatch.html
                    try:
                        arrow = mpatches.FancyArrowPatch(
                            (current_date + pd.Timedelta(hours=12), row_data['median']),
                            (current_date + pd.Timedelta(hours=12), row_data['target_lag']),
                            mutation_scale=20, edgecolor='None', label='normalization correction')
                        ax.add_patch(arrow)
                        make_label = False
                    except:
                        pass

                else:
                    try:
                        ax.broken_barh([xrange], yrange, facecolors='#FDD835', alpha=1, edgecolor='None')
                        ax.broken_barh([xrange], yrange_target, facecolors='#8BC34A', alpha=1, edgecolor='None')
                        try:
                            arrow = mpatches.FancyArrowPatch(
                                (current_date + pd.Timedelta(hours=12), row_data['median']),
                                (current_date + pd.Timedelta(hours=12), row_data['target_lag']),
                                mutation_scale=20, edgecolor='None')
                            ax.add_patch(arrow)
                        except:
                            pass
                    except:
                        pass

            # # Recommended default lag search winsize after normalization
            # recommended_upper = overlay_default_df['target_lag'] \
            #                     + overlay_default_df['recommended_default_winsize']
            # recommended_lower = overlay_default_df['target_lag'] \
            #                     - overlay_default_df['recommended_default_winsize']
            # ax.axhline(recommended_upper.iloc[0])
            # ax.axhline(recommended_lower.iloc[0])

        else:
            ax.text(0.5, 0.5, "No high-quality lags found, lag normalization failed",
                    horizontalalignment='center', verticalalignment='center',
                    transform=ax.transAxes,
                    size=20, color='white', backgroundcolor='red', zorder=100)

    def get(self):
        return self.lgs_winsize, self.new_iteration_data

    def loop_segments(self,
                      data_df: pd.DataFrame,
                      file_idx: pd.Timestamp,
                      filename: str,
                      segment_lagtimes_df: pd.DataFrame):
        """
        Loop through all data segments in each file

        For example, one six hour raw data file consists of twelve half-hour segments.

        Parameters
        ----------
        data_df: pandas DataFrame
            Contains file data.
        file_idx: pandas Timestamp
            Date and time of the current file.
        filename: str
            The file name of the file that contains the current segment.
        segment_lagtimes_df: pandas DataFrame
            Contains lag time info for each segment. Needed as attribute to
            add results of each iteration to results from all previous iterations.
            DataFrame is empty during iteration 1 and filled with the first results
            after iteration 1 has finished.

        Returns
        -------
        pandas DataFrame with lag time for each segment.

        """
        counter_segment = -1
        data_df['index'] = pd.to_datetime(data_df.index)

        # Loop segments
        segment_grouped = data_df.groupby(pd.Grouper(key='index', freq=self.lgs_segment_dur))
        for segment_key, segment_df in segment_grouped:
            counter_segment += 1
            segment_name = f"{segment_df.index[0].strftime('%Y%m%d%H%M%S')}_iter{self.iteration}"
            segment_start = segment_df.index[0]
            segment_end = segment_df.index[-1]
            self.logger.info(f"    Working on FILE: {filename}    SEGMENT: {segment_name}    "
                             f"Lag search window: {self.lgs_winsize}    Step-size: {self.shift_stepsize}    "
                             f"segment start: {segment_start}    segment end: {segment_end}")

            # # todo expand segment df (soon)
            # # Extend segment
            # this_segment_end = segment_df.index[-1]
            # segment_extension_end = this_segment_end + pd.to_timedelta(self.data_segment_overhang)
            # extension_range_df = data_df.loc[this_segment_end:segment_extension_end]
            # segment_df = segment_df.append(extension_range_df[1:])  # Remove 1st row, overlaps w/ segment
            # print(f"Adding segment overhang {self.data_segment_overhang}, new end time {segment_df.index[-1]} ... ")

            # Search lag
            cov_df, props_peak_auto = \
                lag.LagSearch(loop_instance=self,
                              segment_df=segment_df,
                              segment_name=segment_name,
                              segment_start=segment_start,
                              segment_end=segment_end,
                              outdir_plots=self.outdirs[f'{self.phase}-2_{self.phase_files}_covariances_plots'],
                              outdir_data=self.outdirs[f'{self.phase}-1_{self.phase_files}_covariances'],
                              file_idx=file_idx,
                              filename=filename).get()

            # Collect results
            numvals_var_reference = segment_df[self.var_reference].dropna().size
            numvals_var_lagged = segment_df[self.var_lagged].dropna().size

            segment_lagtimes_df.loc[segment_name, 'file_date'] = file_idx
            segment_lagtimes_df.loc[segment_name, 'start'] = segment_start
            segment_lagtimes_df.loc[segment_name, 'end'] = segment_end
            segment_lagtimes_df.loc[segment_name, f'numvals_{self.var_reference}'] = numvals_var_reference
            segment_lagtimes_df.loc[segment_name, f'numvals_{self.var_lagged}'] = numvals_var_lagged
            segment_lagtimes_df.loc[segment_name, f'lagsearch_start'] = self.lgs_winsize[0]
            segment_lagtimes_df.loc[segment_name, f'lagsearch_end'] = self.lgs_winsize[1]
            segment_lagtimes_df.loc[segment_name, f'iteration'] = self.iteration

            # Store lag results in overview
            stor_dict = {'stor_idx': segment_name,
                         'flag_df': cov_df}
            segment_lagtimes_df = self.store_search_results(stor_df=segment_lagtimes_df,
                                                            stor_cols={'shift': 'PEAK-COVABSMAX_SHIFT',
                                                                       'cov': 'PEAK-COVABSMAX_COV',
                                                                       'timestamp': 'PEAK-COVABSMAX_TIMESTAMP'},
                                                            flag_col='flag_peak_max_cov_abs', **stor_dict)
            segment_lagtimes_df = self.store_search_results(stor_df=segment_lagtimes_df,
                                                            stor_cols={'shift': 'PEAK-AUTO_SHIFT',
                                                                       'cov': 'PEAK-AUTO_COV',
                                                                       'timestamp': 'PEAK-AUTO_TIMESTAMP'},
                                                            flag_col='flag_peak_auto', **stor_dict)
            segment_lagtimes_df = self.store_search_results(stor_df=segment_lagtimes_df,
                                                            stor_cols={'shift': 'DEFAULT-LAG_SHIFT',
                                                                       'cov': 'DEFAULT-LAG_COV',
                                                                       'timestamp': 'DEFAULT-LAG_TIMESTAMP'},
                                                            flag_col='flag_instantaneous_default_lag', **stor_dict)

            # Save found segment lag times after each segment
            outfile = f'{self.iteration}_segments_found_lag_times_after_iteration-{self.iteration}.csv'
            segment_lagtimes_df.to_csv(self.outdirs[f'{self.phase}-3_{self.phase_files}_time_lags_overview'] / outfile)
            # self.logger.info(f"Saved segment lag times in {outfile}")

        return segment_lagtimes_df

    def store_search_results(self, stor_df: pd.DataFrame, stor_cols: dict, stor_idx: int,
                             flag_df: pd.DataFrame, flag_col: str):
        """
        Get flag info from a DataFrame and store info in another DataFrame.

        Parameters
        ----------
        stor_df: pandas DataFrame
            Stores overview info.
        stor_cols: dict
            Contains column names used to store info in stor_df.
        stor_idx: int
        flag_df: pandas DataFrame
            Contains flag info.
        flag_col: str
            Flag column name in flag_df.

        Returns
        -------
        pandas DataFrame

        """
        # Init columns
        for key, col in stor_cols.items():
            stor_df.loc[stor_idx, col] = np.nan

        # Get indices of peaks and instantaneous default lag
        flag_idx = lag.LagSearch.get_peak_idx(df=flag_df, flag_col=flag_col)

        # Get info from flag_df and store it in stor_df
        if flag_idx:
            stor_df.loc[stor_idx, stor_cols['shift']] = flag_df.iloc[flag_idx]['shift']  # todo (maybe) give in sec
            stor_df.loc[stor_idx, stor_cols['cov']] = flag_df.iloc[flag_idx]['cov']
            stor_df.loc[stor_idx, stor_cols['timestamp']] = flag_df.iloc[flag_idx]['index']
        return stor_df

    @staticmethod
    def filter_series(filter_col, filter_equal_to, df, series_col):
        """
        Filter series based on values in another column.

        Parameters
        ----------
        filter_col: str
            Name of column that is used for filtering.
        filter_equal_to: int
            Value that must be contained in filter_col to keep value.
        df: pandas DataFrame
        series_col: str
            Name of column that is filtered.

        Returns
        -------
        pandas Series

        """
        filter_this_iteration = df[filter_col] == filter_equal_to
        series_filtered = df.loc[filter_this_iteration, series_col]
        return series_filtered

    def check_for_prev_results(self, iteration: int):
        """
        Check if previous results exist

        Returns
        -------
        segment_lagtimes_df

        """
        filepath_this_iteration = self.outdirs[f'{self.phase}-3_{self.phase_files}_time_lags_overview'] \
                                  / f'{iteration}_segments_found_lag_times_after_iteration-{iteration}.csv'
        filepath_prev_iteration = self.outdirs[f'{self.phase}-3_{self.phase_files}_time_lags_overview'] \
                                  / f'{iteration - 1}_segments_found_lag_times_after_iteration-{iteration - 1}.csv'
        if os.path.exists(filepath_this_iteration):
            # Check if current iteration was already done previously
            self.logger.info(f"(!) Results for iteration {iteration} already exist in file {filepath_this_iteration}, "
                             f"skipping to next iteration.")
            return None
        else:
            if os.path.exists(filepath_prev_iteration):
                # Get results from previous iteration
                segment_lagtimes_df = files.read_segment_lagtimes_file(filepath=filepath_prev_iteration)

                # Adjusted lagsearch window, from a previous iteration
                lgs_winsize = \
                    [segment_lagtimes_df['lagsearch_next_start'].unique()[0],
                     segment_lagtimes_df['lagsearch_next_end'].unique()[0]]
                self.logger.info(f"    Found results from previous iteration in {filepath_prev_iteration}, "
                                 f"time window for lag search set to {lgs_winsize}")
            else:
                # No results from previous iteration found
                segment_lagtimes_df = pd.DataFrame()
                lgs_winsize = self.lgs_winsize  # Initial lagsearch window
                self.logger.info(f"Initial time window for lag search set to {lgs_winsize}")

            return segment_lagtimes_df, lgs_winsize

    def add_lagsearch_adj_info(self, segment_lagtimes_df, iteration, next_lgs_winsize):
        """Add adjusted lagsearch window for next iteration to CSV"""
        segment_lagtimes_df.loc[:, 'lagsearch_next_start'] = next_lgs_winsize[0]
        segment_lagtimes_df.loc[:, 'lagsearch_next_end'] = next_lgs_winsize[1]

        # Save found segment lag times with next lagsearch info after all files
        filename_segments = f'{iteration}_segments_found_lag_times_after_iteration-{iteration}.csv'
        segment_lagtimes_df.to_csv(
            self.outdirs[f'{self.phase}-3_{self.phase_files}_time_lags_overview'] / filename_segments)
        return

    @staticmethod
    def lagsearch_settings(lgs_winsize: list, force_min_stepsize: bool = False):
        """
        Set step size of shift during lag search and calculate number of
        histogram bins.

        Parameters
        ----------
        lgs_winsize: list
            Contains start record for lag search in [0] and end record
            in [1], e.g. [-1000, 1000] searches between records -1000 and
            +1000.

        force_min_stepsize: bool
            If True, sets the stepsize for lag search to 1, otherwise
            stepsize is calculated from win_lagsearch.


        Returns
        -------
        shift_stepsize: int
            Step-size (number of records) for lag search.
        hist_bin_range: range
            Bin range for the histogram of found lag times. The histogram
            is used to narrow down win_lagsearch.

        Examples
        --------
        >> shift_stepsize, hist_bin_range = lagsearch_settings(win_lagsearch=-1000, 1000])
        >> print(shift_stepsize)
        10
        >> print(hist_bin_range)
        range(-1000, 1000, 50)

        >> shift_stepsize, hist_bin_range = lagsearch_settings(win_lagsearch=-1000, 1000], force_min_stepsize=True)
        >> print(shift_stepsize)
        1
        >> print(hist_bin_range)
        range(-1000, 1000, 50)

        """
        # range_win_lagsearch = np.sum(np.abs(win_lagsearch))
        range_lgs_winsize = np.abs(lgs_winsize[0] - lgs_winsize[1])

        if not force_min_stepsize:
            shift_stepsize = int(range_lgs_winsize / 200)
            # shift_stepsize = int(range_win_lagsearch / 100 / 2)
            shift_stepsize = 1 if shift_stepsize < 1 else shift_stepsize  # Step-size cannot be less than 1
        else:
            shift_stepsize = 1

        # if shift_stepsize != 1:
        #     bins_stepsize = int(shift_stepsize * 5)
        # else:
        #     bins_stepsize = int(shift_stepsize)
        max_num_bins = range_lgs_winsize / shift_stepsize
        max_allowed_bins = int(max_num_bins / 3)
        bins_stepsize = int(range_lgs_winsize / max_allowed_bins)

        hist_bin_range = range(int(lgs_winsize[0]), int(lgs_winsize[1]), bins_stepsize)

        # if shift_stepsize != 1:
        #     bins_stepsize = int(range_win_lagsearch / (shift_stepsize * 5))
        #     hist_bin_range = range(int(win_lagsearch[0]), int(win_lagsearch[1]), bins_stepsize)
        #     # hist_bin_range = range(int(win_lagsearch[0]), int(win_lagsearch[1]), int(shift_stepsize * 5))
        # else:
        #     bins_stepsize = shift_stepsize
        #     hist_bin_range = range(int(win_lagsearch[0]), int(win_lagsearch[1]), bins_stepsize)

        return shift_stepsize, hist_bin_range


class PlotLoopResults:
    def __init__(self, dyco_instance, plot_cov_collection=True,
                 plot_hist=True, plot_timeseries_segment_lagtimes=True):
        self.dyco_instance = dyco_instance
        self.outdirs = dyco_instance.outdirs
        self.lgs_num_iter = dyco_instance.lgs_num_iter
        self.lgs_hist_perc_thres = dyco_instance.lgs_hist_perc_thres
        self.plot_cov_collection = plot_cov_collection
        self.plot_hist = plot_hist
        self.plot_timeseries_segment_lagtimes = plot_timeseries_segment_lagtimes
        self.phase = dyco_instance.phase
        self.phase_files = dyco_instance.phase_files
        self.new_iteration_data = dyco_instance.new_iteration_data

        self.logger = setup_dyco.create_logger(logfile_path=dyco_instance.logfile_path, name=__name__)

    def run(self):
        # # In case no new iteration data was created, there is no need to plot data
        # if not self.new_iteration_data:
        #     return None

        # Covariance collection
        if self.plot_cov_collection:
            # Plot covariances from lag search for each segment in one plot
            self.cov_collection(indir=self.outdirs[f'{self.phase}-1_{self.phase_files}_covariances'],
                                outdir=self.outdirs[f'{self.phase}-2_{self.phase_files}_covariances_plots'],
                                logfile_path=self.dyco_instance.logfile_path)
            # self.logger.info(f"Saved covariance collection plot in {outfile}")

        # Histogram
        if self.plot_hist:
            # Read found lag time results from very last iteration
            segment_lagtimes_df = files.read_segment_lagtimes_file(
                filepath=self.outdirs[f'{self.phase}-3_{self.phase_files}_time_lags_overview'] /
                         f'{self.lgs_num_iter}_segments_found_lag_times_after_iteration-{self.lgs_num_iter}.csv')

            # Set search window for lag, depending on iteration
            hist_series = Loop.filter_series(filter_col='iteration', filter_equal_to=self.lgs_num_iter,
                                             df=segment_lagtimes_df, series_col='PEAK-COVABSMAX_SHIFT')
            last_lgs_winsize = [int(segment_lagtimes_df.iloc[-1]['lagsearch_start']),
                                int(segment_lagtimes_df.iloc[-1]['lagsearch_end'])]
            _, hist_num_bins = Loop.lagsearch_settings(lgs_winsize=last_lgs_winsize)
            _ = lag.AdjustLagsearchWindow(series=hist_series,
                                          outdir=self.outdirs[
                                              f'{self.phase}-4_{self.phase_files}_time_lags_overview_histograms'],
                                          iteration=self.lgs_num_iter,
                                          plot=True,
                                          hist_num_bins=hist_num_bins,
                                          remove_fringe_bins=False,
                                          perc_threshold=self.lgs_hist_perc_thres).get()
            self.logger.info(f"Created histogram plot for lag search range {last_lgs_winsize}")

        # Timeseries of lag times
        if self.plot_timeseries_segment_lagtimes:
            # Read found lag time results from very last iteration
            segment_lagtimes_df = files.read_segment_lagtimes_file(
                filepath=self.outdirs[f'{self.phase}-3_{self.phase_files}_time_lags_overview']
                         / f'{self.lgs_num_iter}_segments_found_lag_times_after_iteration-{self.lgs_num_iter}.csv')
            outfile = Loop.plot_segment_lagtimes_ts(segment_lagtimes_df=segment_lagtimes_df,
                                                    outdir=self.outdirs[
                                                        f'{self.phase}-5_{self.phase_files}_time_lags_overview_timeseries'],
                                                    iteration=self.lgs_num_iter,
                                                    show_all=True,
                                                    phase=self.phase)
            self.logger.info(f"Created time series plot of {len(segment_lagtimes_df)} segments "
                             f"across {self.lgs_num_iter} iterations")

    @staticmethod
    def cov_collection(indir, outdir, logfile_path):
        """
        Read and plot segment covariance files

        Parameters
        ----------
        indir: Path
        outdir: Path

        Returns
        -------
        None

        """

        logger = setup_dyco.create_logger(logfile_path=logfile_path, name=__name__)

        # Figure setup
        gs = gridspec.GridSpec(3, 1)  # rows, cols
        gs.update(wspace=0.3, hspace=0.2, left=0.03, right=0.97, top=0.95, bottom=0.03)
        fig = plt.Figure(facecolor='white', figsize=(16, 12))
        ax1 = fig.add_subplot(gs[0, 0])
        ax2 = fig.add_subplot(gs[1, 0], sharex=ax1)
        ax3 = fig.add_subplot(gs[2, 0], sharex=ax1)

        # Read results from last iteration
        cov_collection_df = pd.DataFrame()
        filelist = os.listdir(str(indir))
        num_files = len(filelist)
        logger.info(f"Plotting covariance collection from {num_files} files")
        for idx, filename in enumerate(filelist):
            # print(idx)
            # if idx > 100:
            #     break
            # print(f"Reading segment covariance file {idx + 1} of {num_files}: {filename}")
            filepath = os.path.join(str(indir), filename)
            segment_cov_df = files.read_segment_lagtimes_file(filepath=filepath)
            cov_collection_df = cov_collection_df.append(segment_cov_df)  # Collect for median and quantiles calc

            # Plot each segment covariance file
            args = dict(alpha=0.05, c='black', lw=0.5, marker='None', zorder=98)
            ax1.plot(segment_cov_df['shift'], segment_cov_df['cov'], **args)
            ax2.plot(segment_cov_df['shift'], segment_cov_df['cov_abs'], **args)
            ax3.plot(segment_cov_df['shift'], segment_cov_df['cov_abs'].divide(segment_cov_df['cov_abs'].max()), **args)

        # Median and quantile lines
        def q25(x):
            return x.quantile(0.25)

        def q75(x):
            return x.quantile(0.75)

        f = {'median', q25, q75}
        labels = ['median', 'q25', 'q75']
        linestyle = ['-', '--', '--']
        colors = ['#f44336', '#2196F3', '#3F51B5']
        iter_df = cov_collection_df[cov_collection_df['segment_name'].str.contains('iter')]
        _df = iter_df.groupby('shift').agg(f)
        args = dict(alpha=1, lw=1, marker='None', zorder=98)
        for label_idx, label_name in enumerate(labels):
            ax1.plot(_df.index, _df['cov'][label_name],
                     label=label_name, ls=linestyle[label_idx], c=colors[label_idx], **args)
            ax2.plot(_df.index, _df['cov_abs'][label_name],
                     label=label_name, ls=linestyle[label_idx], c=colors[label_idx], **args)
            if label_name == 'median':
                series = _df['cov_abs'][label_name].divide(_df['cov_abs'][label_name].max())
                ax3.plot(_df.index, series,
                         label=label_name, ls=linestyle[label_idx], c=colors[label_idx], **args)

        # Format
        ax1.set_ylim(cov_collection_df['cov'].quantile(.1), cov_collection_df['cov'].quantile(.9))
        ax2.set_ylim(cov_collection_df['cov_abs'].quantile(.1), cov_collection_df['cov_abs'].quantile(.9))

        fig.suptitle("Results for all segments and from all iterations")
        text_args = dict(horizontalalignment='left', verticalalignment='top',
                         size=12, color='black', backgroundcolor='none', zorder=100)
        ax1.text(0.02, 0.97, "Covariances",
                 transform=ax1.transAxes, **text_args)
        ax2.text(0.02, 0.97, "Absolute covariances",
                 transform=ax2.transAxes, **text_args)
        ax3.text(0.02, 0.97, "Normalized absolute covariances\n(normalized to max of median line)",
                 transform=ax3.transAxes, **text_args)

        ax1.legend(frameon=False, loc='upper right')
        plot.default_format(ax=ax1, txt_xlabel='', txt_ylabel='covariance', txt_ylabel_units='')
        plot.default_format(ax=ax2, txt_xlabel='', txt_ylabel='absolute covariance', txt_ylabel_units='')
        plot.default_format(ax=ax3, txt_xlabel='shift [records]', txt_ylabel='normalized absolute covariance',
                            txt_ylabel_units='')

        # Save
        outfile = '1_covariance_collection_all_segments.png'
        outpath = outdir / outfile
        fig.savefig(f"{outpath}", format='png', bbox_inches='tight', facecolor='w',
                    transparent=True, dpi=150)



"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020  holukas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import sys

import matplotlib.dates as mdates
import numpy as np
import pandas as pd

from . import files, loop, plot, setup_dyco


class AnalyzeLags:
    """
    Analyze lag search results and create look-up table for lag-time normalization

    * Creates LUT for daily default time lags (aggregated) needed in Phase 1 and Phase 2.
    * Creates LUT for instantaneous time lags needed in Phase 3.

    """

    def __init__(self,
                 dyco_instance,
                 direct_path_to_segment_lagtimes_file=None):

        self.lgs_num_iter = dyco_instance.lgs_num_iter
        self.outdirs = dyco_instance.outdirs
        self.target_lag = dyco_instance.target_lag
        self.direct_path_to_segment_lagtimes_file = direct_path_to_segment_lagtimes_file
        self.phase = dyco_instance.phase
        self.phase_files = dyco_instance.phase_files

        self.logger = setup_dyco.create_logger(logfile_path=dyco_instance.logfile_path, name=__name__)

        self.run()

    def run(self):
        self.lut_lag_times_df, self.lut_available = self.generate_lut_time_lags()

        if self.outdirs:
            if self.phase != 3:
                self.save_lut(lut=self.lut_lag_times_df,
                              outdir=self.outdirs[f'{self.phase}-6_{self.phase_files}_normalization_lookup_table'],
                              outfile='LUT_default_agg_time_lags')
                self.plot_segment_lagtimes_with_agg_default()
            if self.phase == 3:
                self.save_lut(lut=self.lut_lag_times_df,
                              outdir=self.outdirs[f'{self.phase}-6_{self.phase_files}_final_time_lags_lookup_table'],
                              outfile='LUT_final_time_lags')
                outdir = self.outdirs[f'{self.phase}-6_{self.phase_files}_final_time_lags_lookup_table']
                AnalyzeLags.plot_final_instantaneous_lagtimes(
                    outdir=outdir,
                    phase=self.phase,
                    df=files.read_segment_lagtimes_file(filepath=outdir / 'LUT_final_time_lags.csv'))

    # def get(self):
    #     return self.lut_default_lag_times_df, self.lut_available

    @staticmethod
    def plot_final_instantaneous_lagtimes(outdir, phase, df):
        """Read and plot final lag search result: the instantaneous time lags"""

        # Get data
        lagsearch_start = int(df['LAGSEARCH_START'].unique()[0])
        lagsearch_end = int(df['LAGSEARCH_END'].unique()[0])
        abs_limit = int(df['ABS_LIMIT'].unique()[0])

        # Plot
        gs, fig, ax = plot.setup_fig_ax()

        # Accepted reference lags
        ax.plot_date(pd.to_datetime(df.index), df['INSTANTANEOUS_LAG'],
                     alpha=1, marker='o', ms=6, color='black', lw=0, ls='-',
                     label=f'final reference time lag (absolute limit {abs_limit})', markeredgecolor='None', zorder=100)

        # Found lags in Phase 3
        ax.plot_date(pd.to_datetime(df.index), df['PEAK-COVABSMAX_SHIFT'],
                     alpha=1, marker='o', ms=12, color='#FFC107', lw=0, ls='-', markeredgecolor='None', zorder=99,
                     label=f'found Phase 3 time lag (search between {lagsearch_start} and {lagsearch_end})')

        # Marks lags that were outside limit and therefore set to the default lag
        set_to_default_lags = df.loc[df['SET_TO_DEFAULT'] == True, ['INSTANTANEOUS_LAG']]
        ax.plot_date(pd.to_datetime(set_to_default_lags.index), set_to_default_lags['INSTANTANEOUS_LAG'],
                     alpha=1, marker='o', ms=12, color='#8BC34A', lw=0, ls='-', markeredgecolor='None', zorder=99,
                     label=f'found Phase 3 time lag was set to default')

        plot.default_format(ax=ax, label_color='black', fontsize=12,
                            txt_xlabel='segment date', txt_ylabel='lag', txt_ylabel_units='[records]')

        txt_info = f"PHASE {phase}: FINAL REFERENCE TIME LAGS"
        font = {'family': 'sans-serif', 'color': 'black', 'weight': 'bold', 'size': 20, 'alpha': 1, }
        ax.set_title(txt_info, fontdict=font)
        # ax.text(0.02, 0.98, txt_info, horizontalalignment='left', verticalalignment='top',
        #         transform=ax.transAxes, backgroundcolor='none', zorder=100, fontdict=font)

        ax.axhline(0, color='black', ls='-', lw=1, label='default lag', zorder=98)
        limit = df['ABS_LIMIT'].unique()[0]
        ax.axhline(limit, color='#d32f2f', ls='--', lw=1, label='upper lag acceptance limit', zorder=98)
        ax.axhline(limit * -1, color='#7B1FA2', ls='--', lw=1, label='lower lag acceptance limit', zorder=98)
        font = {'family': 'sans-serif', 'size': 10}
        ax.legend(frameon=True, loc='upper right', prop=font).set_zorder(100)

        # Automatic tick locations and formats
        locator = mdates.AutoDateLocator(minticks=5, maxticks=20)
        formatter = mdates.ConciseDateFormatter(locator)
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter(formatter)

        # Save
        outpath = outdir / f'PHASE-{phase}_FINAL_TIME_LAGS_FOR_REFERENCE_VAR'
        # print(f"Saving time series of found segment lag times in {outpath} ...")
        fig.savefig(f"{outpath}.png", format='png', bbox_inches='tight', facecolor='w',
                    transparent=True, dpi=150)

    def plot_segment_lagtimes_with_agg_default(self):
        """
        Read and plot found time lags from all or only very last iteration(s),
        used in Phase 1 and Phase 2.
        """
        segment_lagtimes_df = files.read_segment_lagtimes_file(
            filepath=self.outdirs[f'{self.phase}-3_{self.phase_files}_time_lags_overview']
                     / f'{self.lgs_num_iter}_segments_found_lag_times_after_iteration-{self.lgs_num_iter}.csv')
        loop.Loop.plot_segment_lagtimes_ts(segment_lagtimes_df=segment_lagtimes_df,
                                           outdir=self.outdirs[
                                               f'{self.phase}-6_{self.phase_files}_normalization_lookup_table'],
                                           iteration=self.lgs_num_iter,
                                           show_all=False,
                                           overlay_default=True,
                                           overlay_default_df=self.lut_lag_times_df,
                                           overlay_target_val=self.target_lag,
                                           phase=self.phase)

    def save_lut(self, lut, outdir, outfile):
        outpath = outdir / outfile
        lut.to_csv(f"{outpath}.csv")

    def generate_lut_time_lags(self):
        """Generate LUT based on found time lags from last iteration"""

        # Load results from last iteration
        last_iteration = self.lgs_num_iter
        if self.outdirs:
            filepath_last_iteration = self.outdirs[f'{self.phase}-3_{self.phase_files}_time_lags_overview'] \
                                      / f'{last_iteration}_segments_found_lag_times_after_iteration-{last_iteration}.csv'

        else:
            filepath_last_iteration = self.direct_path_to_segment_lagtimes_file  # Implemented for unittest

        segment_lagtimes_last_iteration_df = self.filter_dataframe(filter_col='iteration',
                                                                   filter_equal_to=last_iteration,
                                                                   df=files.read_segment_lagtimes_file(
                                                                       filepath=filepath_last_iteration))

        if self.phase != 3:
            # LUT for Phase 1 or Phase 2, aggregated daily lags
            lut_df, lut_available = self.make_lut_agg(target_lag=self.target_lag,
                                                      segment_lagtimes_df=segment_lagtimes_last_iteration_df)
        else:
            # LUT for Phase 3, instantaneous lags
            lut_df, lut_available = self.make_lut_instantaneous(segment_lagtimes_df=segment_lagtimes_last_iteration_df,
                                                                default_lag=self.target_lag)

        if lut_available:
            self.logger.info(f"Finished creating look-up table for default lag times and normalization correction")
        else:
            self.logger.critical(f"(!) Look-up Table for default lag times and normalization correction is empty, "
                                 f"stopping script.")
            sys.exit()

        return lut_df, lut_available

    @staticmethod
    def filter_dataframe(filter_col, filter_equal_to, df):
        filter_this_iteration = df[filter_col] == filter_equal_to
        df_filtered = df.loc[filter_this_iteration, :]
        return df_filtered

    def make_lut_instantaneous(self, segment_lagtimes_df: pd.DataFrame, default_lag: int):
        """
        Generate instantaneous lag look-up table that contains the found lag time
        for each averaging interval

        Used in Phase 3.

        Parameters
        ----------
        default_lag: int
            The median of high-quality peaks is moved to target_lag.

        Returns
        -------
        pandas DataFrame with default lag times for each day

        """
        # Prepare df
        _segment_lagtimes_df = segment_lagtimes_df.set_index('file_date')
        lut_df = _segment_lagtimes_df[['PEAK-COVABSMAX_SHIFT']].copy()  # file_date needed for look-up
        # lut_df = segment_lagtimes_df[['file_date', 'PEAK-COVABSMAX_SHIFT']].copy()  # file_date needed for look-up
        # lut_df.set_index('file_date', inplace=True)
        lut_df['LAGSEARCH_START'] = np.nan
        lut_df['LAGSEARCH_END'] = np.nan
        lut_df['ABS_LIMIT'] = np.nan
        lut_df['DEFAULT_LAG'] = np.nan
        lut_df['SET_TO_DEFAULT'] = np.nan
        lut_df['INSTANTANEOUS_LAG'] = np.nan
        lut_df['DEFAULT_LAG'] = default_lag

        lut_df['ABS_LIMIT'] = 50
        lut_df['LAGSEARCH_START'] = _segment_lagtimes_df['lagsearch_start']
        lut_df['LAGSEARCH_END'] = _segment_lagtimes_df['lagsearch_end']

        # Replace found lags above absolute threshold with default lag, keep others
        filter_set_to_default = (lut_df['PEAK-COVABSMAX_SHIFT'].abs() > lut_df['ABS_LIMIT']) | \
                                (lut_df['PEAK-COVABSMAX_SHIFT'].isnull())
        lut_df['SET_TO_DEFAULT'] = filter_set_to_default
        lut_df.loc[filter_set_to_default, 'INSTANTANEOUS_LAG'] = lut_df.loc[filter_set_to_default, 'DEFAULT_LAG']
        lut_df.loc[~filter_set_to_default, 'INSTANTANEOUS_LAG'] = lut_df.loc[
            ~filter_set_to_default, 'PEAK-COVABSMAX_SHIFT']

        self.logger.info(f"Created look-up table for {len(lut_df.index)} dates")
        self.logger.info(f"    First date: {lut_df.index[0]}    Last date: {lut_df.index[-1]}")

        # Check for gaps
        missing_df = self.check_missing(df=lut_df,
                                        col='INSTANTANEOUS_LAG')
        if missing_df.empty:
            # All lags available
            pass
        else:
            # Some or all lags missing, fill with default lag
            self.logger.warning(f"No lag was available for dates: {missing_df.index.to_list()}")
            self.logger.warning(f"Filling missing lags with default lag, affected dates: {missing_df.index.to_list()}")
            lut_df['INSTANTANEOUS_LAG'].fillna(lut_df['DEFAULT_LAG'])

        lut_available = True
        return lut_df, lut_available

    def make_lut_agg(self, target_lag: int, segment_lagtimes_df: pd.DataFrame):
        """
        Generate aggregated look-up table that contains the default lag time for each day

        Default lag times are determined by
            (1) pooling data of the current day with data of the day before and
                the day after,
            (2) calculating the median of the pooled data.

        Used in Phase 1 and Phase 2.

        Parameters
        ----------
        target_lag: int
            The median of high-quality peaks is moved to target_lag.

        Returns
        -------
        pandas DataFrame with default lag times for each day

        """
        lut_df = pd.DataFrame()
        peaks_hq_S = self.get_hq_peaks(df=segment_lagtimes_df)  # High-quality covariance peaks

        if peaks_hq_S.empty:
            lut_available = False
            return lut_df, lut_available

        unique_dates = np.unique(peaks_hq_S.index.date)
        for this_date in unique_dates:
            from_date = this_date - pd.Timedelta('1D')
            to_date = this_date + pd.Timedelta('1D')
            filter_around_this_day = (peaks_hq_S.index.date > from_date) & \
                                     (peaks_hq_S.index.date <= to_date)
            subset = peaks_hq_S[filter_around_this_day]
            num_vals = len(subset)

            if num_vals >= 5:
                # print(f"{this_date}    {num_vals}    {subset.median()}")
                lut_df.loc[this_date, 'median'] = subset.median()
                lut_df.loc[this_date, 'counts'] = subset.count()
                lut_df.loc[this_date, 'from'] = from_date
                lut_df.loc[this_date, 'to'] = to_date
            else:
                lut_df.loc[this_date, 'median'] = np.nan
                lut_df.loc[this_date, 'counts'] = subset.count()
                lut_df.loc[this_date, 'from'] = from_date
                lut_df.loc[this_date, 'to'] = to_date

        lut_df['target_lag'] = target_lag
        lut_df['correction'] = -1 * (lut_df['target_lag'] - lut_df['median'])

        self.logger.info(f"Created look-up table for {len(lut_df.index)} dates")
        self.logger.info(f"    First date: {lut_df.index[0]}    Last date: {lut_df.index[-1]}")

        # Fill gaps in 'correction'
        missing_df = self.check_missing(df=lut_df,
                                        col='correction')
        self.logger.warning(f"No correction could be generated from data for dates: {missing_df.index.to_list()}")
        self.logger.warning(f"Filling missing corrections for dates: {missing_df.index.to_list()}")
        lut_df['correction'].fillna(method='ffill', inplace=True, limit=1)

        try:
            lut_df['recommended_default_winsize'] = \
                int((np.abs(lut_df['correction'].min() - lut_df['correction'].max())) / 2) * 1.2
        except ValueError:
            lut_df['recommended_default_winsize'] = np.nan

        lut_available = True
        return lut_df, lut_available

    def check_missing(self, df, col):
        filter_missing = df[col].isnull()
        missing_df = df[filter_missing]
        return missing_df

    def get_hq_peaks(self, df):
        """
        Detect high-quality covariance peaks in results from last lag search iteration

        High-quality means that during the covariance calculations the max covariance
        peak and the automatically detected peak yielded the same results, i.e. the
        same record.

        Parameters
        ----------
        df: pandas DataFrame containing results from the last lag search iteration

        Returns
        -------
        pandas Series of high-quality lag times, given as number of records

        """
        df.set_index('start', inplace=True)
        df.index = pd.to_datetime(df.index)
        peaks_hq_S = df.loc[df['PEAK-COVABSMAX_SHIFT'] == df['PEAK-AUTO_SHIFT'],
                            'PEAK-COVABSMAX_SHIFT']
        peaks_hq_S.index = peaks_hq_S.index.to_pydatetime()  # Convert to DatetimeIndex
        return peaks_hq_S

# def find_default(self, df):
#     plot_df = df[['cov_max_shift']].copy()
#
#     for b in range(1, 4, 1):
#         bins = 2
#         plot_df['group'] = pd.cut(plot_df['cov_max_shift'],
#                                   bins=bins, retbins=False,
#                                   duplicates='drop', labels=False)
#         plot_df_agg = plot_df.groupby('group').agg(['count', 'min', 'max'])
#         idxmax = plot_df_agg['cov_max_shift']['count'].idxmax()  # Most counts
#         group_max = plot_df_agg.iloc[idxmax].name
#
#         plot_df_agg['count_maxperc'] = \
#             plot_df_agg['cov_max_shift']['count'] / plot_df_agg['cov_max_shift']['count'].sum()
#         # plot_df_agg['cov_max_shift']['count'] / plot_df_agg.iloc[idxmax]['cov_max_shift']['count']
#
#         plot_df = plot_df.loc[plot_df['group'] == group_max]
#
#     median = plot_df['cov_max_shift'].median()
#     _min = plot_df['cov_max_shift'].min()
#     _max = plot_df['cov_max_shift'].max()
#
#     print(plot_df)
#     print(f"{median}  {_min}  {_max}")



"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020  holukas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import datetime as dt

import numpy as np
import pandas as pd


def read_segment_lagtimes_file(filepath):
    """
    Read file.

    Reading segment covariances and lag search results for each segment.
    Can be used for all text files for which the .read_csv args are valid.

    Parameters
    ----------
    filepath: str

    Returns
    -------
    pandas DataFrame

    """
    # parse = lambda x: dt.datetime.strptime(x, '%Y%m%d%H%M%S')
    found_lags_df = pd.read_csv(filepath,
                                skiprows=None,
                                header=0,
                                # names=header_cols_list,
                                # na_values=-9999,
                                encoding='utf-8',
                                delimiter=',',
                                mangle_dupe_cols=True,
                                # keep_date_col=False,
                                parse_dates=False,
                                # date_parser=parse,
                                index_col=0,
                                dtype=None,
                                engine='c')
    return found_lags_df


def read_raw_data(filepath, data_timestamp_format):
    header_rows_list = [0]
    skip_rows_list = []
    header_section_rows = [0]

    num_data_cols = \
        length_data_cols(filepath=filepath,
                         header_rows_list=header_rows_list,
                         skip_rows_list=skip_rows_list)

    num_header_cols, header_cols_df = \
        length_header_cols(filepath=filepath,
                           header_rows_list=header_rows_list,
                           skip_rows_list=skip_rows_list)

    more_data_cols_than_header_cols, num_missing_header_cols = \
        data_vs_header(num_data_cols=num_data_cols,
                       num_header_cols=num_header_cols)

    header_cols_list = \
        generate_missing_cols(header_cols_df=header_cols_df,
                              more_data_cols_than_header_cols=more_data_cols_than_header_cols,
                              num_missing_header_cols=num_missing_header_cols)

    if data_timestamp_format:
        parse = lambda x: dt.datetime.strptime(x, data_timestamp_format)
        date_parser = parse
        parse_dates = True
        index_col = 0
    else:
        date_parser = None
        parse_dates = False
        index_col = None

    data_df = pd.read_csv(filepath,
                          skiprows=header_section_rows,
                          header=None,
                          names=header_cols_list,
                          na_values=-9999,
                          encoding='utf-8',
                          delimiter=',',
                          mangle_dupe_cols=True,
                          keep_date_col=False,
                          parse_dates=parse_dates,
                          date_parser=date_parser,
                          index_col=index_col,
                          dtype=None,
                          engine='c',
                          nrows=None)

    return data_df


def calc_true_resolution(num_records, data_nominal_res, expected_records, expected_duration):
    ratio = num_records / expected_records
    if (ratio > 0.999) and (ratio < 1.001):
        # file_complete = True
        true_resolution = np.float64(expected_duration / num_records)
    else:
        # file_complete = False
        true_resolution = data_nominal_res
    return true_resolution


def insert_timestamp(df, file_info_row, num_records, data_nominal_res, expected_records, expected_duration):
    true_resolution = calc_true_resolution(num_records=num_records, data_nominal_res=data_nominal_res,
                                           expected_records=expected_records, expected_duration=expected_duration)
    df['sec'] = df.index * true_resolution
    df['file_start_dt'] = file_info_row['start']
    df['TIMESTAMP'] = pd.to_datetime(df['file_start_dt']) \
                      + pd.to_timedelta(df['sec'], unit='s')
    df.drop(['sec', 'file_start_dt'], axis=1, inplace=True)
    df.set_index('TIMESTAMP', inplace=True)

    return df


def add_data_stats(df, true_resolution, filename, files_overview_df, found_records, fnm_date_format):
    # Detect overall frequency
    data_duration = found_records * true_resolution
    data_freq = np.float64(found_records / data_duration)

    idx = dt.datetime.strptime(filename, fnm_date_format)  # Use filename datetime info as index

    files_overview_df.loc[idx, 'first_record'] = df.index[0]
    files_overview_df.loc[idx, 'last_record'] = df.index[-1]
    files_overview_df.loc[idx, 'file_duration'] = (df.index[-1] - df.index[0]).total_seconds()
    files_overview_df.loc[idx, 'found_records'] = found_records
    files_overview_df.loc[idx, 'data_freq'] = data_freq

    return files_overview_df


def generate_missing_cols(header_cols_df, more_data_cols_than_header_cols, num_missing_header_cols):
    # Generate missing header columns if necessary
    header_cols_list = header_cols_df.columns.to_list()
    generated_missing_header_cols_list = []
    if more_data_cols_than_header_cols:
        for m in list(range(1, num_missing_header_cols + 1)):
            missing_col = (f'unknown_{m}')
            generated_missing_header_cols_list.append(missing_col)
            header_cols_list.append(missing_col)
    return header_cols_list


def length_data_cols(filepath, header_rows_list, skip_rows_list):
    # Check number of columns of the first data row after the header part
    skip_num_lines = len(header_rows_list) + len(skip_rows_list)
    first_data_row_df = pd.read_csv(filepath,
                                    skiprows=skip_num_lines,
                                    header=None,
                                    nrows=1)
    return first_data_row_df.columns.size


def length_header_cols(filepath, header_rows_list, skip_rows_list):
    # Check number of columns of the header part
    header_cols_df = pd.read_csv(filepath,
                                 skiprows=skip_rows_list,
                                 header=header_rows_list,
                                 nrows=0)
    return header_cols_df.columns.size, header_cols_df


def data_vs_header(num_data_cols, num_header_cols):
    # Check if there are more data columns than header columns
    if num_data_cols > num_header_cols:
        more_data_cols_than_header_cols = True
        num_missing_header_cols = num_data_cols - num_header_cols
    else:
        more_data_cols_than_header_cols = False
        num_missing_header_cols = 0
    return more_data_cols_than_header_cols, num_missing_header_cols


def collect_file_data(self, data_df, file_idx, data_collection_df):
    if file_idx == self.files_overview_df.index[0]:
        data_collection_df = data_df.copy()
    else:
        data_collection_df = data_collection_df.append(data_df)
    return data_collection_df



"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020  holukas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.collections import LineCollection
from scipy.signal import find_peaks

from . import plot, setup_dyco


class LagSearch:
    """
    Determine the time lag for each file by calculating covariances
    and finding covariance peaks
    """

    def __init__(self,
                 loop_instance,
                 segment_df: pd.DataFrame,
                 segment_name: str,
                 file_idx: pd.Timestamp,
                 segment_start: pd.Timestamp,
                 segment_end: pd.Timestamp,
                 filename: str,
                 outdir_data: Path = None,
                 outdir_plots: Path = None):

        self.var_reference = loop_instance.var_reference
        self.var_lagged = loop_instance.var_lagged
        self.lgs_winsize = loop_instance.lgs_winsize
        self.iteration = loop_instance.iteration
        self.shift_stepsize = loop_instance.shift_stepsize  # Negative moves lagged values "upwards" in column
        self.logfile_path = loop_instance.logfile_path
        self.phase = loop_instance.phase
        self.phase_files = loop_instance.phase_files

        self.segment_df = segment_df
        self.segment_name = segment_name
        self.segment_start = segment_start
        self.segment_end = segment_end
        self.outdir_data = outdir_data
        self.outdir_plots = outdir_plots
        self.file_idx = file_idx
        self.filename = filename

        if self.logfile_path:
            self.logger = setup_dyco.create_logger(logfile_path=self.logfile_path, name=__name__)

        self.run()

    def run(self):
        """Execute processing stack"""
        self.cov_df = self.setup_lagsearch_df(lgs_winsize=self.lgs_winsize,
                                              shift_stepsize=self.shift_stepsize,
                                              segment_name=self.segment_name)

        # Detect and flag max covariance peak
        self.cov_df = self.find_max_cov_peak(segment_df=self.segment_df,
                                             cov_df=self.cov_df,
                                             var_reference=self.var_reference,
                                             var_lagged=self.var_lagged)

        # Detect and flag automatic peak
        self.cov_df, self.props_peak_auto = self.find_auto_peak(cov_df=self.cov_df)

        # Detect and flag instantaneous default lag in Phase 3
        self.cov_df = self.flag_instantaneous_default_lag(default=0,
                                                          cov_df=self.cov_df)

        if self.outdir_data:
            self.save_cov_data(cov_df=self.cov_df)

        # Get indices of peaks and instantaneous default lag
        self.idx_peak_cov_abs_max = \
            self.get_peak_idx(df=self.cov_df, flag_col='flag_peak_max_cov_abs')
        self.idx_peak_auto = \
            self.get_peak_idx(df=self.cov_df, flag_col='flag_peak_auto')
        self.idx_instantaneous_default_lag = \
            self.get_peak_idx(df=self.cov_df, flag_col='flag_instantaneous_default_lag')

        # Plot covariance
        fig = self.make_scatter_cov()
        if self.outdir_plots:
            self.save_cov_plot(fig=fig)
        return None

    def flag_instantaneous_default_lag(self, default, cov_df):
        """Flags location with instantaneous default time lag"""
        if not self.phase == 3:
            return cov_df
        default_ix = cov_df['shift'].loc[cov_df['shift'] == default].index[0]
        cov_df.loc[default_ix, 'flag_instantaneous_default_lag'] = True
        return cov_df

    @staticmethod
    def setup_lagsearch_df(lgs_winsize: list, shift_stepsize: int, segment_name):
        df = pd.DataFrame(columns=['index', 'segment_name', 'shift', 'cov', 'cov_abs',
                                   'flag_peak_max_cov_abs', 'flag_peak_auto'])
        df['shift'] = range(int(lgs_winsize[0]),
                            int(lgs_winsize[1]) + shift_stepsize,
                            shift_stepsize)
        df['index'] = np.nan
        df['segment_name'] = segment_name
        df['cov'] = np.nan
        df['cov_abs'] = np.nan
        df['flag_peak_max_cov_abs'] = False  # Flag True = found peak
        df['flag_peak_auto'] = False
        df['flag_instantaneous_default_lag'] = False
        return df

    @staticmethod
    def find_auto_peak(cov_df: pd.DataFrame):
        """
        Automatically find peaks in covariance time series

        The found peak is flagged TRUE in cov_df.

        Peaks are searched automatically using scipy's .find_peaks method.
        The peak_score is calculated for each peak, the top-scoring peaks
        are retained. If the previously calculated max covariance peak is
        part of the top-scoring peaks, the record at which these two peaks
        were found is flagged in cov_df. Basically, this method works as a
        validation step to detect high-quality covariance peaks that are
        later used to calculate default lag times.


        Parameters
        ----------
        cov_df: pandas DataFrame
            Contains covariances for each shift.

        Returns
        -------
        cov_df
        props_peak_df

        """

        found_peaks_idx, found_peaks_dict = find_peaks(cov_df['cov_abs'],
                                                       height=0, width=0, prominence=0)
        found_peaks_props_df = pd.DataFrame.from_dict(found_peaks_dict)
        found_peaks_props_df['idx_in_cov_df'] = found_peaks_idx

        # Calculate peak score, a combination of peak_height, prominences and width_heights
        found_peaks_props_df['peak_score'] = found_peaks_props_df['prominences'] \
                                             * found_peaks_props_df['width_heights'] \
                                             * found_peaks_props_df['peak_heights']
        found_peaks_props_df['peak_score'] = found_peaks_props_df['peak_score'] ** .5  # Make numbers smaller
        found_peaks_props_df['peak_rank'] = found_peaks_props_df['peak_score'].rank(ascending=False)

        score_threshold = found_peaks_props_df['peak_score'].quantile(0.9)
        top_scoring_peaks_df = found_peaks_props_df.loc[found_peaks_props_df['peak_score'] >= score_threshold]
        top_scoring_peaks_df = top_scoring_peaks_df.sort_values(by=['peak_score', 'prominences', 'width_heights'],
                                                                ascending=False)

        idx_peak_cov_abs_max = \
            LagSearch.get_peak_idx(df=cov_df, flag_col='flag_peak_max_cov_abs')

        # Check if peak of max absolute covariance is also in auto-detected peaks
        if idx_peak_cov_abs_max in top_scoring_peaks_df['idx_in_cov_df'].values:
            props_peak_df = top_scoring_peaks_df.iloc[
                top_scoring_peaks_df['idx_in_cov_df'].values == idx_peak_cov_abs_max]
            props_peak_df = props_peak_df.iloc[0]
            peak_idx = int(props_peak_df['idx_in_cov_df'])
            cov_df.loc[peak_idx, 'flag_peak_auto'] = True
        else:
            props_peak_df = pd.DataFrame()

        # props_peak_df = pd.DataFrame()  # Props for one peak
        # if len(found_peaks_idx) > 0:
        #     max_width_height_idx = found_peaks_dict['width_heights'].argmax()  # Good metric to find peaks
        #     most_prominent_idx = found_peaks_dict['prominences'].argmax()
        #
        #     if max_width_height_idx == most_prominent_idx:
        #         peak_idx = found_peaks_idx[max_width_height_idx]
        #         cov_df.loc[peak_idx, 'flag_peak_auto'] = True
        #         props_peak_df = found_peaks_props_df.iloc[max_width_height_idx]

        return cov_df, props_peak_df

    def get(self):
        return self.cov_df, self.props_peak_auto

    @staticmethod
    def find_max_cov_peak(segment_df, cov_df, var_lagged, var_reference):
        """
        Find maximum absolute covariance
        """

        _segment_df = segment_df.copy()
        _segment_df['index'] = _segment_df.index

        # Check if data column is empty
        if _segment_df[var_lagged].dropna().empty:
            pass

        else:
            for ix, row in cov_df.iterrows():
                shift = int(row['shift'])
                try:
                    if shift < 0:
                        index_shifted = str(_segment_df['index'][-shift])  # Note the negative sign
                    else:
                        index_shifted = pd.NaT
                    scalar_data_shifted = _segment_df[var_lagged].shift(shift)
                    # cov = _segment_df[ref_sig].corr(scalar_data_shifted)
                    cov = _segment_df[var_reference].cov(scalar_data_shifted)
                    cov_df.loc[cov_df['shift'] == row['shift'], 'cov'] = cov
                    cov_df.loc[cov_df['shift'] == row['shift'], 'index'] = index_shifted

                except IndexError:
                    # If not enough data in the file to perform the shift, continue
                    # to the next shift and try again. This can happen for the last
                    # segments in each file, when there is no more data available
                    # at the end.
                    continue

            # Results
            cov_df['cov_abs'] = cov_df['cov'].abs()
            cov_max_ix = cov_df['cov_abs'].idxmax()
            # cov_max_shift = lagsearch_df.iloc[cov_max_ix]['shift']
            # cov_max = lagsearch_df.iloc[cov_max_ix]['cov']
            # cov_max_timestamp = lagsearch_df.iloc[cov_max_ix]['index']
            cov_df.loc[cov_max_ix, 'flag_peak_max_cov_abs'] = True

        return cov_df

    def save_cov_data(self, cov_df):
        outfile = f'{self.segment_name}_segment_covariance_iteration-{self.iteration}.csv'
        outpath = self.outdir_data / outfile
        cov_df.to_csv(f"{outpath}")
        # self.logger.info(f"Saved covariance data in {outfile}")
        return None

    @staticmethod
    def get_peak_idx(df, flag_col):
        """
        Search a boolean column for *True* and return index if successful

        Parameters
        ----------
        df: pandas DataFrame
        flag_col: str
            Column name in *df*.

        Returns
        -------
        The index where flag_col is *True*.

        """
        # Peak of maximum covariance
        if True in df[flag_col].values:
            idx = df.loc[df[flag_col] == True, :].index.values[0]
        else:
            idx = False
        return idx

        # # Peak of maximum covariance
        # if True in df['flag_peak_max_cov_abs'].values:
        #     idx_peak_cov_abs_max = df.loc[df['flag_peak_max_cov_abs'] == True, :].index.values[0]
        # else:
        #     idx_peak_cov_abs_max = False
        #
        # # Auto-detected peak
        # if True in df['flag_peak_auto'].values:
        #     idx_peak_auto = df.loc[df['flag_peak_auto'] == True, :].index.values[0]
        # else:
        #     idx_peak_auto = False
        #
        # # Location of the instantaneous default lag
        # if True in df['flag_instantaneous_default_lag'].values:
        #     idx_instantaneous_default_lag = df.loc[df['flag_instantaneous_default_lag'] == True, :].index.values[0]
        # else:
        #     idx_instantaneous_default_lag = False
        #
        # return idx_peak_cov_abs_max, idx_peak_auto, idx_instantaneous_default_lag

    def save_cov_plot(self, fig):
        """Save covariance plot for segment"""
        outfile = f'{self.segment_name}_segment_covariance_iteration-{self.iteration}.png'
        outpath = self.outdir_plots / outfile
        fig.savefig(f"{outpath}", format='png', bbox_inches='tight', facecolor='w',
                    transparent=True, dpi=100)
        # self.logger.info(f"Saved covariance plot in {outfile}")
        return

    def make_scatter_cov(self):
        """Make scatter plot with z-values as colors and display found max covariance."""

        gs, fig, ax = plot.setup_fig_ax()

        # Covariance and shift data
        x_shift = self.cov_df.loc[:, 'shift']
        y_cov = self.cov_df.loc[:, 'cov']
        z_cov_abs = self.cov_df.loc[:, 'cov_abs']

        # Main plot: covariances per shift, vals from abs cov as scatter point colors
        ax.scatter(x_shift, y_cov, c=z_cov_abs,
                   alpha=0.9, edgecolors='none',
                   marker='o', s=24, cmap='coolwarm', zorder=98)

        # Use abs cov also as line colors
        self.z_as_colored_lines(fig=fig, ax=ax,
                                x=x_shift,
                                y=y_cov,
                                z=z_cov_abs)

        txt_info = \
            f"PHASE: {self.phase}\n" \
            f"Iteration: {self.iteration}\n" \
            f"Time lag search window: from {self.lgs_winsize[0]} to {self.lgs_winsize[1]} records\n" \
            f"Segment name: {self.segment_name}\n" \
            f"Segment start: {self.segment_start}\n" \
            f"Segment end: {self.segment_end}\n" \
            f"File: {self.filename} - File date: {self.file_idx}\n" \
            f"Lag search step size: {self.shift_stepsize} records\n"

        # Markers for points of interest, e.g. peaks
        txt_info = self.mark_max_cov_abs_peak(ax=ax, txt_info=txt_info)
        txt_info = self.mark_auto_detected_peak(ax=ax, txt_info=txt_info)
        txt_info = self.mark_instantaneous_default_lag(ax=ax, txt_info=txt_info)

        # Add info text
        ax.text(0.02, 0.98, txt_info,
                horizontalalignment='left', verticalalignment='top',
                transform=ax.transAxes, size=10, color='black', backgroundcolor='none', zorder=100)

        # Format & legend
        plot.default_format(ax=ax, label_color='black', fontsize=12,
                            txt_xlabel='lag [records]', txt_ylabel='covariance', txt_ylabel_units='-')
        ax.legend(frameon=False, loc='upper right').set_zorder(100)

        return fig

    def z_as_colored_lines(self, fig, ax, x, y, z):
        # z values as colors
        # From: https://matplotlib.org/3.1.1/gallery/lines_bars_and_markers/multicolored_line.html
        # Create a set of line segments so that we can color them individually
        # This creates the points as a N x 1 x 2 array so that we can stack points
        # together easily to get the segments. The segments array for line collection
        # needs to be (numlines) x (points per line) x 2 (for x and y)
        points = np.array([x, y]).T.reshape(-1, 1, 2)
        segments = np.concatenate([points[:-1], points[1:]], axis=1)

        # Create a continuous norm to map from data points to colors
        norm = plt.Normalize(z.min(), z.max())
        lc = LineCollection(segments, cmap='coolwarm', norm=norm)
        # Set the values used for colormapping
        lc.set_array(z)
        lc.set_linewidth(2)
        line = ax.add_collection(lc)
        cbar = fig.colorbar(line, ax=ax)
        cbar.set_label('absolute covariance', rotation=90)

    def mark_max_cov_abs_peak(self, ax, txt_info):
        """Mark peak of max absolute covariance"""
        if self.idx_peak_cov_abs_max:
            ax.scatter(self.cov_df.iloc[self.idx_peak_cov_abs_max]['shift'],
                       self.cov_df.iloc[self.idx_peak_cov_abs_max]['cov'],
                       alpha=1, edgecolors='red', marker='o', s=72, c='red',
                       label='maximum absolute covariance', zorder=99)
            txt_info += \
                f"\nFOUND PEAK MAX ABS COV\n" \
                f"    cov {self.cov_df.iloc[self.idx_peak_cov_abs_max]['cov']:.3f}\n" \
                f"    record {self.cov_df.iloc[self.idx_peak_cov_abs_max]['shift']}\n"
        else:
            txt_info += \
                f"\n(!)NO PEAK MAX ABS COV FOUND\n"
        return txt_info

    def mark_instantaneous_default_lag(self, ax, txt_info):
        """Mark instantaneous default time lag, used in Phase 3"""
        if self.idx_instantaneous_default_lag:
            ax.scatter(self.cov_df.iloc[self.idx_instantaneous_default_lag]['shift'],
                       self.cov_df.iloc[self.idx_instantaneous_default_lag]['cov'],
                       alpha=1, edgecolors='green', marker='o', s=200, c='None',
                       label='instantaneous default lag', zorder=90)
            txt_info += \
                f"\nTIME LAG SET TO DEFAULT (PHASE 3)\n" \
                f"    lag time was set to default\n" \
                f"    cov {self.cov_df.iloc[self.idx_instantaneous_default_lag]['cov']:.3f}\n" \
                f"    record {self.cov_df.iloc[self.idx_instantaneous_default_lag]['shift']}\n"
        # else:
        #     txt_info += \
        #         f"\n(!)NO AUTO-PEAK FOUND\n"
        return txt_info

    def mark_auto_detected_peak(self, ax, txt_info):
        """Mark auto-detected peak"""
        if self.idx_peak_auto:
            ax.scatter(self.cov_df.iloc[self.idx_peak_auto]['shift'],
                       self.cov_df.iloc[self.idx_peak_auto]['cov'],
                       alpha=1, edgecolors='black', marker='o', s=200, c='None',
                       label='auto-detected peak', zorder=90)
            txt_info += \
                f"\nFOUND AUTO-PEAK\n" \
                f"    cov {self.cov_df.iloc[self.idx_peak_auto]['cov']:.3f}\n" \
                f"    record {self.cov_df.iloc[self.idx_peak_auto]['shift']}\n" \
                f"    peak_score {self.props_peak_auto['peak_score']:.0f}\n" \
                f"    peak_rank {self.props_peak_auto['peak_rank']:.0f}\n" \
                f"    peak_height {self.props_peak_auto['peak_heights']:.0f}\n" \
                f"    prominence {self.props_peak_auto['prominences']:.0f}\n" \
                f"    width {self.props_peak_auto['widths']:.0f}\n" \
                f"    width_height {self.props_peak_auto['width_heights']:.0f}\n"
        else:
            txt_info += \
                f"\n(!)NO AUTO-PEAK FOUND\n"
        return txt_info


class AdjustLagsearchWindow():
    def __init__(self, series, iteration, plot=True, hist_num_bins=30, remove_fringe_bins=True,
                 perc_threshold=0.9, outdir=None):
        self.series = series.dropna()  # NaNs yield error in histogram
        self.numvals_series = self.series.size
        self.outdir = outdir
        self.iteration = iteration
        self.plot = plot
        self.hist_num_bins = hist_num_bins
        self.remove_fringe_bins = remove_fringe_bins
        self.perc_threshold = perc_threshold

        self.run()

    def run(self):
        self.lgs_winsize_adj, self.peak_max_count_idx, self.start_idx, self.end_idx, \
        self.counts, self.divisions, self.peak_most_prom_idx = \
            self.find_hist_peaks()

        if self.plot:
            self.plot_results_hist(hist_bins=self.divisions)

    def find_hist_peaks(self):
        """Find peak in histogram of found lag times."""

        # Make histogram of found lag times, remove fringe bins at start and end
        counts, divisions = self.calc_hist(series=self.series, bins=self.hist_num_bins,
                                           remove_fringe_bins=self.remove_fringe_bins)

        # Search bin with most found lag times
        peak_max_count_idx = self.search_bin_max_counts(counts=counts)

        # Search most prominent bin
        peak_most_prom_idx = self.search_bin_most_prominent(counts=counts)

        # Adjust lag search time window for next iteration
        lgs_winsize_adj, start_idx, end_idx = self.adjust_lgs_winsize(counts=counts, divisions=divisions,
                                                                      perc_threshold=self.perc_threshold,
                                                                      peak_max_count_idx=peak_max_count_idx,
                                                                      peak_most_prom_idx=peak_most_prom_idx)

        # Check if most prominent peak is also the max peak
        if peak_most_prom_idx in peak_max_count_idx:  # todo (maybe)
            clear_peak_idx = np.where(peak_max_count_idx == peak_most_prom_idx)
        else:
            clear_peak_idx = False

        lgs_winsize_adj = [divisions[start_idx], divisions[end_idx]]
        lgs_winsize_adj = [int(x) for x in
                           lgs_winsize_adj]  # Convert elements in array to integers, needed for indexing

        return lgs_winsize_adj, peak_max_count_idx, start_idx, end_idx, counts, divisions, peak_most_prom_idx

    def get(self):
        return self.lgs_winsize_adj

    def search_bin_max_counts(self, counts):
        # print("Searching peak of maximum counts ...")
        max_count = np.amax(counts)
        peak_max_count_idx = np.where(counts == np.amax(max_count))  # Returns array in tuple
        if len(peak_max_count_idx) == 1:
            peak_max_count_idx = peak_max_count_idx[0]  # Yields array of index or multiple indices
        return peak_max_count_idx

    @staticmethod
    def calc_hist(series=False, bins=20, remove_fringe_bins=False):
        """Calculate histogram of found lag times."""
        counts, divisions = np.histogram(series, bins=bins)
        # Remove first and last bins from histogram. In case of unclear lag times
        # data tend to accumulate in these edge regions of the search window.
        if remove_fringe_bins and len(counts) >= 5:
            counts = counts[1:-1]
            divisions = divisions[1:-1]  # Contains start values of bins
        return counts, divisions

    @staticmethod
    def search_bin_most_prominent(counts):
        # kudos: https://www.kaggle.com/simongrest/finding-peaks-in-the-histograms-of-the-variables
        # Increase prominence until only one single peak is found
        # print("Searching most prominent peak ...")
        peak_most_prom_idx = []
        prom = 0  # Prominence for peak finding
        while (len(peak_most_prom_idx) == 0) or (len(peak_most_prom_idx) > 1):
            prom += 1
            if prom > 40:
                peak_most_prom_idx = False
                break
            peak_most_prom_idx, props = find_peaks(counts, prominence=prom)
            # print(f"Prominence: {prom}    Peaks at: {peak_most_prom_idx}")
        if peak_most_prom_idx:
            peak_most_prom_idx = int(peak_most_prom_idx)
        return peak_most_prom_idx

    def adjust_lgs_winsize(self, counts, divisions, perc_threshold, peak_max_count_idx, peak_most_prom_idx):
        """Set new time window for next lag search, based on previous results.

        Includes more and more bins around the bin where most lag times were found
        until a threshold is reached.
        """
        start_idx, end_idx = self.include_bins_next_to_peak(peak_max_count_idx=peak_max_count_idx,
                                                            peak_most_prom_idx=peak_most_prom_idx)

        counts_total = np.sum(counts)
        perc = 0
        while perc < perc_threshold:
            start_idx = start_idx - 1 if start_idx > 0 else start_idx
            end_idx = end_idx + 1 if end_idx < len(counts) else end_idx
            c = counts[start_idx:end_idx]
            perc = np.sum(c) / counts_total
            # print(f"Expanding lag window: {perc}  from record: {start_idx}  to record: {end_idx}")
            if (start_idx == 0) and (end_idx == len(counts)):
                break
        lgs_winsize_adj = [divisions[start_idx], divisions[end_idx]]
        lgs_winsize_adj = [int(x) for x in
                           lgs_winsize_adj]  # Convert elements in array to integers, needed for indexing
        return lgs_winsize_adj, start_idx, end_idx

    def include_bins_next_to_peak(self, peak_max_count_idx, peak_most_prom_idx):
        """Include histogram bins next to the bin for which max was found and the
        most prominent bin.

        Since multiple max count peaks can be detected in the histogram, all found
        peaks are considered and all bins before and after each detected peak are
        included to calculate the adjusted start and end indices.

        For examples:
            Three peaks were with max count were found in the histogram. The peaks
            were found in bins 5, 9 and 14:
                peak_max_count_index = [5,9,14]
            The most prominent peak was detected in bin 2:
                peak_most_prom_idx = 2
            Then the bins before the max count peaks are included:
                start_idx = [4,5,8,9,13,14]
            Then the bins after the max count peaks are included:
                end_idx = [4,5,6,8,9,10,13,14,15]
            Then the max count peaks are combined with the most prominent peak,
            using np.unique() in case of overlapping bins:
                start_end_idx = [2,4,5,6,8,9,10,13,14,15]
            The start_idx is the min of this collection:
                start_idx = 2
            The end_idx is the max of this collection:
                end_idx = 15
            The adjusted time window for lag search starts with the starting time
            of bin 2 and ends with the end time with bin 15 (starting time is added
            in next steps).
        """
        start_idx = np.subtract(peak_max_count_idx, 1)  # Include bins before each peak
        start_idx[start_idx < 0] = 0  # Negative index not possible
        end_idx = np.add(peak_max_count_idx, 1)  # Include bins after each peak
        start_end_idx = np.unique(np.concatenate([start_idx, end_idx, [peak_most_prom_idx]]))  # Combine peaks
        start_idx = np.min(start_end_idx)
        end_idx = np.max(start_end_idx[-1])
        return start_idx, end_idx

    def plot_results_hist(self, hist_bins):

        gs, fig, ax = plot.setup_fig_ax()

        # Counts
        # bar_positions = plot_df_agg.index + 0.5  # Shift by half position
        bar_width = (hist_bins[1] - hist_bins[0]) * 0.9  # Calculate bar width
        args = dict(width=bar_width, align='edge')
        ax.bar(x=hist_bins[0:-1], height=self.counts, label='counts', zorder=90, color='#78909c', **args)
        ax.set_xlim(hist_bins[0], hist_bins[-1])

        # Text counts
        for i, v in enumerate(self.counts):
            if v > 0:
                ax.text(hist_bins[0:-1][i] + (bar_width / 2), v, str(v), zorder=99, size=6)

        ax.bar(x=hist_bins[self.peak_max_count_idx], height=self.counts[self.peak_max_count_idx],
               label='most counts', zorder=98, edgecolor='#ef5350', linewidth=4,
               color='None', alpha=0.9, linestyle='-', **args)

        if self.peak_most_prom_idx:
            ax.bar(x=hist_bins[self.peak_most_prom_idx], height=self.counts[self.peak_most_prom_idx],
                   label='most prominent counts peak', zorder=99, edgecolor='#FFA726', linewidth=2,
                   color='None', alpha=0.9, linestyle='--', **args)

        # xtick_labels_int = [int(l) for l in xtick_labels]
        # ax.set_xticks(xtick_labels_int)  # Position of ticks

        ax.axvline(x=hist_bins[self.start_idx], ls='--', c='#42A5F5',
                   label='lag search window start for next iteration')
        ax.axvline(x=hist_bins[self.end_idx], ls='--', c='#AB47BC',
                   label='lag search window end for next iteration')

        txt_info = \
            f"Histogram of found lag times in iteration {self.iteration}\n" \
            f"Number of found lag times: {self.numvals_series}"

        if self.remove_fringe_bins:
            txt_info += "\nFringe bins removed: yes"
        else:
            txt_info += "\nFringe bins removed: no"

        ax.text(0.02, 0.98, txt_info,
                horizontalalignment='left', verticalalignment='top',
                transform=ax.transAxes, size=14, color='black',
                backgroundcolor='none', zorder=100)

        ax.legend(loc='upper right', prop={'size': 10}).set_zorder(100)

        plot.default_format(ax=ax, label_color='black', fontsize=12,
                            txt_xlabel='lag [records]', txt_ylabel='counts', txt_ylabel_units='[#]')

        if self.outdir:

            if self.remove_fringe_bins:
                outfile = f'{self.iteration}_HISTOGRAM_segment_lag_times_iteration-{self.iteration}.png'
            else:
                outfile = f'HISTOGRAM_segment_lag_times_FINAL.png'

            outpath = self.outdir / outfile
            fig.savefig(f"{outpath}", format='png', bbox_inches='tight', facecolor='w',
                        transparent=True, dpi=150)

        # ax.set_xticklabels(division)  # Labels of ticks, shown in plot
        # if len(label_bin_start) > 30:
        #     ax.tick_params(rotation=45)
        # else:
        #     ax.tick_params(rotation=0)
        # default_format(ax=ax, txt_xlabel=f'bin {col[1]}', txt_ylabel=f'{col[0]} counts',
        #                              txt_ylabel_units='[#]')
        # gui.plotfuncs.default_grid(ax=ax)

        return None



"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020  holukas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import datetime as dt
import os
from pathlib import Path

import matplotlib.dates as mdates
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import pandas as pd

from . import analyze, files, loop, setup_dyco


class SummaryPlots():
    def __init__(self, instance_phase_1, instance_phase_2, instance_phase_3):
        self.instance_phase_1 = instance_phase_1
        self.instance_phase_2 = instance_phase_2
        self.instance_phase_3 = instance_phase_3

        self.outdir_summary = self._make_outdir()

        self.logger = setup_dyco.create_logger(logfile_path=instance_phase_3.logfile_path, name=__name__)

        self.run()

    def run(self):
        self.logger.info("CREATING SUMMARY ...")
        collection_df = pd.DataFrame()

        # Plot and collect time lags and covariances from Phase 1 and Phase 2
        self.logger.info("Plotting found time lags, Phases 1-3")
        phase_instances = [self.instance_phase_1, self.instance_phase_2]
        for phase_instance in phase_instances:
            self.summary_plot_segment_lagtimes(phase=phase_instance.phase,
                                               phase_files=phase_instance.phase_files,
                                               outdirs=phase_instance.outdirs,
                                               last_iteration=phase_instance.lgs_num_iter,
                                               target_lag=phase_instance.target_lag)
            collection_df = self._collect_peak_covariances(phase=phase_instance.phase,
                                                           phase_files=phase_instance.phase_files,
                                                           outdirs=phase_instance.outdirs,
                                                           last_iteration=phase_instance.lgs_num_iter,
                                                           collection_df=collection_df)

        # Phase 3
        phase_instances = [self.instance_phase_3]
        for phase_instance in phase_instances:
            self.summary_plot_instantaneous_lagtimes(phase=phase_instance.phase,
                                                     phase_files=phase_instance.phase_files,
                                                     outdirs=phase_instance.outdirs)
            collection_df, shift_colname_out = self._collect_instantaneous_lags(phase=phase_instance.phase,
                                                                                phase_files=phase_instance.phase_files,
                                                                                outdirs=phase_instance.outdirs,
                                                                                collection_df=collection_df)
            collection_df = self._collect_instantaneous_covariances(phase=phase_instance.phase,
                                                                    phase_files=phase_instance.phase_files,
                                                                    outdirs=phase_instance.outdirs,
                                                                    collection_df=collection_df,
                                                                    shift_colname_out=shift_colname_out)

        self.summary_plot_covariances(collection_df=collection_df,
                                      outdir=self.outdir_summary,
                                      last_iter_phase_1=self.instance_phase_1.lgs_num_iter,
                                      last_iter_phase_2=self.instance_phase_2.lgs_num_iter)

        self.logger.info("CREATING SUMMARY ... Done.")
        self.logger.info("Finished DYCO processing.")

    def _add_lines_cov(self, ax, df, last_iter_phase_1, last_iter_phase_2):
        props = {'y': df[f'PHASE-1_ITER-{last_iter_phase_1}_PEAK-COVABSMAX_COV'],
                 'label': f'Phase 1',
                 # 'label': f'PHASE-1_ITER-{last_iter_phase_1}_PEAK-COVABSMAX_COV',
                 'zorder': 100, 'alpha': 1, 'marker': 's', 'ms': 6, 'lw': 1, 'ls': 'dotted',
                 'markeredgecolor': '#80DEEA', 'color': '#80DEEA', 'markerfacecolor': 'white'}
        ax.plot_date(df.index, **props)

        props = {'y': df[f'PHASE-2_ITER-{last_iter_phase_1}_PEAK-COVABSMAX_COV'],
                 'label': f'Phase 2',
                 # 'label': f'PHASE-2_ITER-{last_iter_phase_1}_PEAK-COVABSMAX_COV',
                 'zorder': 100, 'alpha': 1, 'marker': 's', 'ms': 6, 'lw': 1, 'ls': '--',
                 'markeredgecolor': '#F48FB1', 'color': '#F48FB1', 'markerfacecolor': 'white'}
        ax.plot_date(df.index, **props)

        props = {'y': df['PHASE-3_INSTANTANEOUS_LAG_COV'],
                 'label': 'Phase 3',
                 # 'label': 'PHASE-3_INSTANTANEOUS_LAG_COV',
                 'zorder': 100, 'alpha': 1, 'marker': 'o', 'ms': 6, 'lw': 2, 'ls': '-',
                 'markeredgecolor': '#8BC34A', 'color': '#8BC34A', 'markerfacecolor': '#8BC34A'}
        ax.plot_date(df.index, **props)

    def _format_plot(self, ax, title, show_legend=True):
        # Automatic tick locations and formats
        locator = mdates.AutoDateLocator(minticks=5, maxticks=20)
        formatter = mdates.ConciseDateFormatter(locator)
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter(formatter)

        default_format(ax=ax, label_color='black', fontsize=12,
                       txt_xlabel='file date', txt_ylabel='covariance', txt_ylabel_units='')

        font = {'family': 'sans-serif', 'color': 'black', 'weight': 'bold', 'size': 12, 'alpha': 1}
        ax.set_title(title, fontdict=font, loc='left', pad=12)

        ax.axhline(0, color='black', ls='-', lw=1, zorder=1)
        font = {'family': 'sans-serif', 'size': 10}
        if show_legend:
            ax.legend(frameon=True, loc='upper right', prop=font).set_zorder(100)

    def summary_plot_covariances(self, collection_df, outdir, last_iter_phase_1, last_iter_phase_2):
        self.logger.info("Plotting covariance evolution, Phases 1-3 ...")
        gs = gridspec.GridSpec(2, 1)  # rows, cols
        gs.update(wspace=0.3, hspace=0.3, left=0.03, right=0.97, top=0.97, bottom=0.03)
        fig = plt.Figure(facecolor='white', figsize=(16, 9))
        ax1 = fig.add_subplot(gs[0, 0])
        ax2 = fig.add_subplot(gs[1, 0])
        props_ax = {'last_iter_phase_1': last_iter_phase_1, 'last_iter_phase_2': last_iter_phase_2}

        # ax1: Covariance evolution
        # =========================
        self._add_lines_cov(ax=ax1, df=collection_df, **props_ax)
        self._format_plot(ax=ax1, title="Half-hourly covariance evolution")

        # ax2: Cumulative
        # ===============
        collection_df_cum = collection_df.cumsum()
        self._add_lines_cov(ax=ax2, df=collection_df_cum, **props_ax)
        self._format_plot(ax=ax2, title="Cumulative covariance evolution", show_legend=False)

        # Save
        outfile = f'summary_covariance_evolution'
        outpath = outdir / outfile
        # print(f"Saving time series of found segment lag times in {outpath} ...")
        fig.savefig(f"{outpath}.png", format='png', bbox_inches='tight', facecolor='w', transparent=True, dpi=150)
        a = 1

    def summary_plot_segment_lagtimes(self, phase: int, phase_files: str, outdirs,
                                      last_iteration: int, target_lag: int):
        """Plot time lags found in last iteration of each Phase incl. aggregated defaults"""
        # Get data
        segment_lagtimes_df = self._get_segment_lagtimes_last_iteration(outdirs=outdirs, phase=phase,
                                                                        phase_files=phase_files,
                                                                        last_iteration=last_iteration)
        lut_df = self._get_lut_last_iteration(outdirs=outdirs, phase=phase, phase_files=phase_files)

        # Make plot
        loop.Loop.plot_segment_lagtimes_ts(segment_lagtimes_df=segment_lagtimes_df,
                                           outdir=self.outdir_summary,
                                           iteration=last_iteration,
                                           show_all=False,
                                           overlay_default=True,
                                           overlay_default_df=lut_df,
                                           overlay_target_val=target_lag,
                                           phase=phase)

    def summary_plot_instantaneous_lagtimes(self, phase, phase_files, outdirs):
        df = self._get_instantaneous_time_lags(phase=phase, phase_files=phase_files, outdirs=outdirs)
        analyze.AnalyzeLags.plot_final_instantaneous_lagtimes(outdir=self.outdir_summary, phase=phase, df=df)

    def _collect_instantaneous_covariances(self, outdirs, phase, phase_files, collection_df, shift_colname_out):
        """Get instantaneous covariances at instantaneous time lags from Phase 3"""
        collection_df.index = pd.to_datetime(collection_df.index)
        indir = outdirs[f'{phase}-1_{phase_files}_covariances']
        filelist = os.listdir(str(indir))
        num_files = len(filelist)
        for idx, filename in enumerate(filelist):
            file_date = dt.datetime.strptime(filename, '%Y%m%d%H%M%S_iter1_segment_covariance_iteration-1.csv')

            # Check if datetime of covariance file is contained in collection
            if file_date in collection_df.index:
                # Read covariance file containing all covariances per shift/lag
                filepath = os.path.join(str(indir), filename)
                cov_df = files.read_segment_lagtimes_file(filepath=filepath)

                # From the collection, get the instantaneous shift/lag that was found for this file
                shift_instantaneous = \
                    collection_df.loc[collection_df.index == file_date, shift_colname_out].values[0]

                # From the covariance file, get the covariance for the instantaneous shift/lag
                filter_instantaneous = cov_df['shift'] == shift_instantaneous
                cov_instantaneous = cov_df.loc[filter_instantaneous, 'cov'].values[0]

                # Store instantaneous covariance in collection
                cov_colname_out = f'PHASE-{phase}_INSTANTANEOUS_LAG_COV'
                collection_df.loc[collection_df.index == file_date, cov_colname_out] = cov_instantaneous
        return collection_df

    def _collect_instantaneous_lags(self, phase, phase_files, outdirs, collection_df):
        df = self._get_instantaneous_time_lags(phase=phase, phase_files=phase_files, outdirs=outdirs)
        df.index = pd.to_datetime(df.index)
        shift_colname_out = f'PHASE-{phase}_INSTANTANEOUS_LAG_SHIFT'
        collection_df[shift_colname_out] = df['INSTANTANEOUS_LAG']
        return collection_df, shift_colname_out

    def _get_instantaneous_time_lags(self, phase, outdirs, phase_files):
        """Get instantaneous time lags from Phase 3 LUT file"""
        indir = outdirs[f'{phase}-6_{phase_files}_final_time_lags_lookup_table']
        df = files.read_segment_lagtimes_file(filepath=indir / 'LUT_final_time_lags.csv')
        return df

    def _collect_peak_covariances(self, outdirs, phase, phase_files, last_iteration,
                                  collection_df):

        segment_lagtimes_df = self._get_segment_lagtimes_last_iteration(outdirs=outdirs, phase=phase,
                                                                        phase_files=phase_files,
                                                                        last_iteration=last_iteration)
        _add_df = segment_lagtimes_df[['file_date', 'PEAK-COVABSMAX_COV', 'PEAK-COVABSMAX_SHIFT']]
        _add_df.set_index('file_date', inplace=True)
        cov_colname_out = f'PHASE-{phase}_ITER-{last_iteration}_PEAK-COVABSMAX_COV'
        shift_colname_out = f'PHASE-{phase}_ITER-{last_iteration}_PEAK-COVABSMAX_SHIFT'

        if phase == 1:
            collection_df = _add_df.copy()
            collection_df.rename(columns={'PEAK-COVABSMAX_COV': cov_colname_out}, inplace=True)
            collection_df.rename(columns={'PEAK-COVABSMAX_SHIFT': shift_colname_out}, inplace=True)
        else:
            collection_df[cov_colname_out] = _add_df['PEAK-COVABSMAX_COV']
            collection_df[shift_colname_out] = _add_df['PEAK-COVABSMAX_SHIFT']

        return collection_df

    def _get_segment_lagtimes_last_iteration(self, outdirs, phase, phase_files, last_iteration):
        filepath = outdirs[f'{phase}-3_{phase_files}_time_lags_overview'] \
                   / f'{last_iteration}_segments_found_lag_times_after_iteration-{last_iteration}.csv'
        df = \
            analyze.AnalyzeLags.filter_dataframe(filter_col='iteration',
                                                 filter_equal_to=last_iteration,
                                                 df=files.read_segment_lagtimes_file(
                                                     filepath=filepath))
        return df

    def _get_lut_last_iteration(self, outdirs, phase, phase_files):
        filepath_lut = outdirs[f'{phase}-6_{phase_files}_normalization_lookup_table'] / 'LUT_default_agg_time_lags.csv'
        lut_df = files.read_segment_lagtimes_file(filepath=filepath_lut)
        return lut_df

    def _make_outdir(self):
        outdir = self.instance_phase_1.outdir / 'SUMMARY'
        if not Path.is_dir(outdir):
            print(f"Creating folder {outdir} ...")
            os.makedirs(outdir)
        return outdir


def limit_data_range_percentiles(df, col, perc_limits):
    p_lower = df[col].quantile(perc_limits[0])
    p_upper = df[col].quantile(perc_limits[1])
    p_filter = (df[col] >= p_lower) & (df[col] <= p_upper)
    df = df[p_filter]
    return df


def default_format(ax, fontsize=12, label_color='black',
                   txt_xlabel='', txt_ylabel='', txt_ylabel_units='',
                   width=1, length=5, direction='in', colors='black', facecolor='white'):
    """Apply default format to plot."""
    ax.set_facecolor(facecolor)
    ax.tick_params(axis='x', width=width, length=length, direction=direction, colors=colors, labelsize=fontsize,
                   top=True)
    ax.tick_params(axis='y', width=width, length=length, direction=direction, colors=colors, labelsize=fontsize,
                   right=True)
    format_spines(ax=ax, color=colors, lw=1)
    if txt_xlabel:
        ax.set_xlabel(txt_xlabel, color=label_color, fontsize=fontsize, fontweight='bold')
    if txt_ylabel and txt_ylabel_units:
        ax.set_ylabel(f'{txt_ylabel}  {txt_ylabel_units}', color=label_color, fontsize=fontsize, fontweight='bold')
    if txt_ylabel and not txt_ylabel_units:
        ax.set_ylabel(f'{txt_ylabel}', color=label_color, fontsize=fontsize, fontweight='bold')


def format_spines(ax, color, lw):
    spines = ['top', 'bottom', 'left', 'right']
    for spine in spines:
        ax.spines[spine].set_color(color)
        ax.spines[spine].set_linewidth(lw)


def setup_fig_ax():
    """Setup grid with one figure and one axis."""
    gs = gridspec.GridSpec(1, 1)  # rows, cols
    gs.update(wspace=0.3, hspace=0.3, left=0.03, right=0.97, top=0.97, bottom=0.03)
    fig = plt.Figure(facecolor='white', figsize=(16, 9))
    ax = fig.add_subplot(gs[0, 0])
    return gs, fig, ax



"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020  holukas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='dyco',
    packages=setuptools.find_packages(),
    # packages=['dyco'],
    version='1.0.6',
    license='GNU General Public License v3 (GPLv3)',
    description='A Python package to detect and compensate for shifting lag times in ecosystem time series',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Lukas Hörtnagl',
    author_email='lukas.hoertnagl@usys.ethz.ch',
    url='https://gitlab.ethz.ch/holukas/dyco-dynamic-lag-compensation',
    download_url='https://pypi.org/project/dyco/',
    keywords=['ecosystem', 'eddy covariance', 'fluxes',
              'time series', 'lag', 'timeshift'],
    install_requires=['pandas==1.0.3', 'matplotlib==3.1.3', 'numpy==1.18.1', 'scipy==1.4.1'],
    classifiers=[
        'Development Status :: 4 - Beta',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



"""
    DYCO Dynamic Lag Compensation
    Copyright (C) 2020  holukas

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import datetime as dt
import time
from pathlib import Path

import numpy as np
import pandas as pd

from . import files, setup_dyco


class RemoveLags:
    """
    Remove time lags: use look-up table to normalize time lags across files
    """

    def __init__(self, dyco_instance):
        self.files_overview_df = dyco_instance.files_overview_df
        self.dat_recs_timestamp_format = dyco_instance.dat_recs_timestamp_format
        self.outdirs = dyco_instance.outdirs
        self.var_target = dyco_instance.var_target
        self.phase = dyco_instance.phase
        self.phase_files = dyco_instance.phase_files
        self.new_iteration_data = dyco_instance.new_iteration_data  # Indicates if new iteration was executed
        self.lgs_num_iter = dyco_instance.lgs_num_iter

        self.logger = setup_dyco.create_logger(logfile_path=dyco_instance.logfile_path, name=__name__)

        if self.phase == 3:

            dir = Path(self.outdirs[f'{self.phase}-6_{self.phase_files}_final_time_lags_lookup_table'])
            file = Path('LUT_final_time_lags.csv')
            self.lut_df = files.read_segment_lagtimes_file(filepath=dir / file)
            self.lut_col = 'INSTANTANEOUS_LAG'
        else:
            # Read default lag times for reference gas
            self.lut_df = self.read_lut_time_lags()
            self.lut_col = 'correction'

        self.run()

    def run(self):
        # If no new iteration data was created, skip normalization
        if not self.new_iteration_data:
            self.logger.warning("")
            self.logger.warning(f"{'*' * 60}")
            self.logger.warning(f"(!) No new iteration data was created in PHASE {self.phase} for {self.phase_files}.")
            self.logger.warning(f"(!) Skipping PHASE {self.phase}, Step 4: no normalized files will be generated.")
            self.logger.warning(f"{'*' * 60}")
            self.logger.warning("")
            return None

        # Loop files
        num_files = self.files_overview_df['file_available'].sum()
        times_needed = []
        files_counter = 0

        for file_idx, file_info_row in self.files_overview_df.iterrows():
            start = time.time()
            txt_info = ""

            # Check file availability
            if file_info_row['file_available'] == 0:
                continue

            if not self.phase == 3:
                this_date = file_info_row['start'].date()
                shift_correction = self.lut_df.loc[this_date][self.lut_col]
            else:
                this_date = file_info_row['start']
                # self.lut_df['file_date'] = pd.to_datetime(self.lut_df['file_date'])

                # If time lags are calculated for segments shorter than the file they
                # are part of, then there are more than one found lag times per file.
                # Therefore, make list of all found time lags and calculate the median.
                # In case segment length is the same as file length and there is only
                # one lag time for this file, then the median has no effect.
                self.lut_df.index = pd.to_datetime(self.lut_df.index)
                shift_correction = self.lut_df.loc[self.lut_df.index == this_date, self.lut_col].to_list()
                shift_correction = np.median(shift_correction)

            if pd.isnull(shift_correction):
                shift_correction = np.nan
                txt_info += f"(!)No lag found in LUT for "
            else:
                # Read and prepare data file
                data_df = files.read_raw_data(filepath=file_info_row['filepath'],
                                              data_timestamp_format=self.dat_recs_timestamp_format)  # nrows for testing

                shift_correction = int(shift_correction)
                data_df = self.shift_var_target(df=data_df,
                                                shift=shift_correction)

                self.save_dyco_files(outdir=self.outdirs[f'{self.phase}-7_{self.phase_files}_normalized'],
                                     original_filename=file_info_row['filename'],
                                     df=data_df,
                                     export_timestamp=True)

            time_needed = time.time() - start
            times_needed.append(time_needed)
            files_counter += 1
            times_needed_mean = np.mean(times_needed)
            remaining_files = num_files - files_counter
            remaining_sec = times_needed_mean * remaining_files
            progress = (files_counter / num_files) * 100
            txt_info += f"File #{files_counter}: {file_info_row['filename']}" \
                        f"    shift correction: {shift_correction}    remaining time: {remaining_sec:.0f}s" \
                        f"    remaining files: {int(remaining_files)}    progress: {progress:.2f}%"
            self.logger.info(txt_info)

    def save_dyco_files(self, df, outdir, original_filename, export_timestamp):
        df.fillna(-9999, inplace=True)
        outpath = outdir / f"{original_filename}_DYCO.csv"
        df.to_csv(outpath, index=export_timestamp)

    def shift_var_target(self, df, shift):
        for col in self.var_target:
            outcol = f"{col}_DYCO"
            df[outcol] = df[col].shift(shift)  # Shift col by found lag
            df.drop([col], axis=1, inplace=True)  # Remove col that was not shifted
        return df

    def read_lut_time_lags(self):
        filepath = self.outdirs[
                       f'{self.phase}-6_{self.phase_files}_normalization_lookup_table'] / f'LUT_default_agg_time_lags.csv'
        parse = lambda x: dt.datetime.strptime(x, '%Y-%m-%d')
        df = pd.read_csv(filepath,
                         skiprows=None,
                         header=0,
                         # names=header_cols_list,
                         # na_values=-9999,
                         encoding='utf-8',
                         delimiter=',',
                         mangle_dupe_cols=True,
                         # keep_date_col=False,
                         parse_dates=True,
                         date_parser=parse,
                         index_col=0,
                         dtype=None,
                         engine='c')
        return df
